import logging
import random
import time
import uuid
from collections import defaultdict, Counter
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

import ray
from ray.air.execution._internal.event_manager import RayEventManager
from ray.air.execution.resources import (
    AcquiredResources,
    ResourceManager,
    ResourceRequest,
)

from ray.air.execution._internal.tracked_actor import TrackedActor
from ray.air.execution._internal.tracked_actor_task import TrackedActorTask
from ray.exceptions import RayTaskError, RayActorError

logger = logging.getLogger(__name__)


class RayActorManager:
    """Management class for Ray actors and actor tasks.

    This class provides an event-based management interface for actors, and
    actor tasks.

    The manager can be used to start actors, stop actors, and schedule and
    track task futures on these actors.
    The manager will then invoke callbacks related to the tracked entities.

    For instance, when an actor is added with
    :meth:`add_actor() <RayActorManager.add_actor>`,
    a :ref:`TrackedActor <ray.air.execution._internal.tracked_actor.TrackedActor`
    object is returned. An ``on_start`` callback can be specified that is invoked
    once the actor successfully started. Similarly, ``on_stop`` and ``on_error``
    can be used to specify callbacks relating to the graceful or ungraceful
    end of an actor's lifetime.

    When scheduling an actor task using
    :meth:`schedule_actor_task()
    <ray.air.execution._internal.actor_manager.RayActorManager.schedule_actor_task>`,
    an ``on_result`` callback can be specified that is invoked when the task
    successfully resolves, and an ``on_error`` callback will resolve when the
    task fails.

    The RayActorManager does not implement any true asynchronous processing. Control
    has to be explicitly yielded to the event manager via :meth:`RayActorManager.next`.
    Callbacks will only be invoked when control is with the RayActorManager, and
    callbacks will always be executed sequentially in order of arriving events.

    Args:
        resource_manager: Resource manager used to request resources for the actors.

    Example:

        .. code-block:: python

            from ray.air.execution import ResourceRequest
            from ray.air.execution._internal import RayActorManager

            actor_manager = RayActorManager()

            # Request an actor
            tracked_actor = actor_manager.add_actor(
                ActorClass,
                kwargs={},
                resource_request=ResourceRequest([{"CPU": 1}]),
                on_start=actor_start_callback,
                on_stop=actor_stop_callback,
                on_error=actor_error_callback
            )

            # Yield control to event manager to start actor
            actor_manager.next()

            # Start task on the actor (ActorClass.foo.remote())
            tracked_actor_task = actor_manager.schedule_actor_task(
                tracked_actor,
                method_name="foo",
                on_result=task_result_callback,
                on_error=task_error_callback
            )

            # Again yield control to event manager to process task futures
            actor_manager.wait()

    """

    def __init__(self, resource_manager: ResourceManager):
        self._resource_manager: ResourceManager = resource_manager

        self._actor_state_events = RayEventManager()
        self._actor_task_events = RayEventManager()

        # ---
        # Tracked actor futures.

        # This maps TrackedActor objects to their futures. We use this to see if an
        # actor has any futures scheduled and to remove them when we terminate an actor.

        # Actors to actor task futures
        self._tracked_actors_to_task_futures: Dict[
            TrackedActor, Set[ray.ObjectRef]
        ] = defaultdict(set)

        # Actors to actor state futures (start/terminate)
        self._tracked_actors_to_state_futures: Dict[
            TrackedActor, Set[ray.ObjectRef]
        ] = defaultdict(set)

        # ---
        # Pending actors.
        # We use three dicts for actors that are requested but not yet started.

        # This dict keeps a list of actors associated with each resource request.
        # We use this to start actors in the correct order when their resources
        # become available.
        self._resource_request_to_pending_actors: Dict[
            ResourceRequest, List[TrackedActor]
        ] = defaultdict(list)

        # This dict stores the actor class, kwargs, and resource request of
        # pending actors. Once the resources are available, we start the remote
        # actor class with its args. We need the resource request to cancel it
        # if needed.
        self._pending_actors_to_attrs: Dict[
            TrackedActor, Tuple[Type, Dict[str, Any], ResourceRequest]
        ] = {}

        # This dict keeps track of cached actor tasks. We can't schedule actor
        # tasks before the actor is actually scheduled/live. So when the caller
        # tries to schedule a task, we cache it here, and schedule it once the
        # actor is started.
        self._pending_actors_to_enqueued_actor_tasks: Dict[
            TrackedActor, List[Tuple[TrackedActorTask, str, Tuple[Any], Dict[str, Any]]]
        ] = defaultdict(list)

        # ---
        # Live actors.
        # We keep one dict for actors that are currently running and a set of
        # actors that we should forcefully kill.

        # This dict associates the TrackedActor object with the Ray actor handle
        # and the resources associated to the actor. We use it to schedule the
        # actual ray tasks, and to return the resources when the actor stopped.
        self._live_actors_to_ray_actors_resources: Dict[
            TrackedActor, Tuple[ray.actor.ActorHandle, AcquiredResources]
        ] = {}
        self._live_resource_cache: Optional[Dict[str, Any]] = None

        # This dict contains all actors that should be killed (after calling
        # `remove_actor()`). Kill requests will be handled in wait().
        self._live_actors_to_kill: Set[TrackedActor] = set()

        # Track failed actors
        self._failed_actor_ids: Set[int] = set()

    def next(self, timeout: Optional[Union[int, float]] = None) -> bool:
        """Yield control to event manager to await the next event and invoke callbacks.

        Calling this method will wait for up to ``timeout`` seconds for the next
        event to arrive.

        When events arrive, callbacks relating to the events will be
        invoked. A timeout of ``None`` will block until the next event arrives.

        Note:
            If an actor task fails with a ``RayActorError``, this is one event,
            but it may trigger _two_ `on_error` callbacks: One for the actor,
            and one for the task.

        Note:
            The ``timeout`` argument is used for pure waiting time for events. It does
            not include time spent on processing callbacks. Depending on the processing
            time of the callbacks, it can take much longer for this function to
            return than the specified timeout.

        Args:
            timeout: Timeout in seconds to wait for next event.

        Returns:
            True if at least one event was processed.

        """
        # First issue any pending forceful actor kills
        actor_killed = self._try_kill_actor()

        # We always try to start actors as this won't trigger an event callback
        self._try_start_actors()

        # If an actor was killed, this was our event, and we return.
        if actor_killed:
            return True

        # Otherwise, collect all futures and await the next.
        resource_futures = self._resource_manager.get_resource_futures()
        actor_state_futures = self._actor_state_events.get_futures()
        actor_task_futures = self._actor_task_events.get_futures()

        # Shuffle state futures
        shuffled_state_futures = list(actor_state_futures)
        random.shuffle(shuffled_state_futures)

        # Shuffle task futures
        shuffled_task_futures = list(actor_task_futures)
        random.shuffle(shuffled_task_futures)

        # Prioritize resource futures over actor state over task futures
        all_futures = resource_futures + shuffled_state_futures + shuffled_task_futures

        start_wait = time.monotonic()
        ready, _ = ray.wait(all_futures, num_returns=1, timeout=timeout)

        if not ready:
            return False

        [future] = ready

        if future in actor_state_futures:
            self._actor_state_events.resolve_future(future)
        elif future in actor_task_futures:
            self._actor_task_events.resolve_future(future)
        else:
            self._handle_ready_resource_future()
            # Ready resource futures don't count as one event as they don't trigger
            # any callbacks. So we repeat until we hit anything that is not a resource
            # future.
            time_taken = time.monotonic() - start_wait
            return self.next(
                timeout=max(1e-9, timeout - time_taken) if timeout is not None else None
            )

        self._try_start_actors()
        return True

    def _actor_start_resolved(self, tracked_actor: TrackedActor, future: ray.ObjectRef):
        """Callback to be invoked when actor started"""
        self._tracked_actors_to_state_futures[tracked_actor].remove(future)

        if tracked_actor._on_start:
            tracked_actor._on_start(tracked_actor)

    def _actor_stop_resolved(self, tracked_actor: TrackedActor):
        """Callback to be invoked when actor stopped"""
        self._cleanup_actor(tracked_actor=tracked_actor)

        if tracked_actor._on_stop:
            tracked_actor._on_stop(tracked_actor)

    def _actor_start_failed(self, tracked_actor: TrackedActor, exception: Exception):
        """Callback to be invoked when actor start/stop failed"""
        self._failed_actor_ids.add(tracked_actor.actor_id)

        self._cleanup_actor(tracked_actor=tracked_actor)

        if tracked_actor._on_error:
            tracked_actor._on_error(tracked_actor, exception)

    def _actor_task_failed(
        self, tracked_actor_task: TrackedActorTask, exception: Exception
    ):
        """Handle an actor task future that became ready.

        - On actor error, trigger actor error callback AND error task error callback
        - On task error, trigger actor task error callback
        - On success, trigger actor task result callback
        """
        tracked_actor = tracked_actor_task._tracked_actor

        if isinstance(exception, RayActorError):
            self._failed_actor_ids.add(tracked_actor.actor_id)

            # Clean up any references to the actor and its futures
            self._cleanup_actor(tracked_actor=tracked_actor)

            # Handle actor state callbacks
            if tracked_actor._on_error:
                tracked_actor._on_error(tracked_actor, exception)

            # Then trigger actor task error callback
            if tracked_actor_task._on_error:
                tracked_actor_task._on_error(tracked_actor, exception)

        elif isinstance(exception, RayTaskError):
            # Otherwise only the task failed. Invoke callback
            if tracked_actor_task._on_error:
                tracked_actor_task._on_error(tracked_actor, exception)
        else:
            raise RuntimeError(
                f"Caught unexpected exception: {exception}"
            ) from exception

    def _actor_task_resolved(self, tracked_actor_task: TrackedActorTask, result: Any):
        tracked_actor = tracked_actor_task._tracked_actor

        # Trigger actor task result callback
        if tracked_actor_task._on_result:
            tracked_actor_task._on_result(tracked_actor, result)

    def _handle_ready_resource_future(self):
        """Handle a resource future that became ready.

        - Update state of the resource manager
        - Try to start one actor
        """
        # Force resource manager to update internal state
        self._resource_manager.update_state()
        # We handle resource futures one by one, so only try to start 1 actor at a time
        self._try_start_actors(max_actors=1)

    def _try_start_actors(self, max_actors: Optional[int] = None) -> int:
        """Try to start up to ``max_actors`` actors.

        This function will iterate through all resource requests we collected for
        pending actors. As long as a resource request can be fulfilled (resources
        are available), we try to start as many actors as possible.

        This will schedule a `Actor.__ray_ready__()` future which, once resolved,
        will trigger the `TrackedActor.on_start` callback.
        """
        started_actors = 0

        # Iterate through all resource requests
        for resource_request in self._resource_request_to_pending_actors:
            if max_actors and started_actors >= max_actors:
                break

            # While we have resources ready and there are actors left to schedule
            while (
                self._resource_manager.has_resources_ready(resource_request)
                and self._resource_request_to_pending_actors[resource_request]
            ):
                # Acquire resources for actor
                acquired_resources = self._resource_manager.acquire_resources(
                    resource_request
                )
                assert acquired_resources

                # Get tracked actor to start
                candidate_actors = self._resource_request_to_pending_actors[
                    resource_request
                ]
                assert candidate_actors

                tracked_actor = candidate_actors.pop(0)

                # Get actor class and arguments
                actor_cls, kwargs, _ = self._pending_actors_to_attrs.pop(tracked_actor)

                if not isinstance(actor_cls, ray.actor.ActorClass):
                    actor_cls = ray.remote(actor_cls)

                # Associate to acquired resources
                [remote_actor_cls] = acquired_resources.annotate_remote_entities(
                    [actor_cls]
                )

                # Start Ray actor
                actor = remote_actor_cls.remote(**kwargs)

                # Track
                self._live_actors_to_ray_actors_resources[tracked_actor] = (
                    actor,
                    acquired_resources,
                )
                self._live_resource_cache = None

                # Schedule ready future
                future = actor.__ray_ready__.remote()

                self._tracked_actors_to_state_futures[tracked_actor].add(future)

                # We need to create the callbacks in a function so tracked_actors
                # are captured correctly.
                def create_callbacks(
                    tracked_actor: TrackedActor, future: ray.ObjectRef
                ):
                    def on_actor_start(result: Any):
                        self._actor_start_resolved(
                            tracked_actor=tracked_actor, future=future
                        )

                    def on_error(exception: Exception):
                        self._actor_start_failed(
                            tracked_actor=tracked_actor, exception=exception
                        )

                    return on_actor_start, on_error

                on_actor_start, on_error = create_callbacks(
                    tracked_actor=tracked_actor, future=future
                )

                self._actor_state_events.track_future(
                    future=future,
                    on_result=on_actor_start,
                    on_error=on_error,
                )

                self._enqueue_cached_actor_tasks(tracked_actor=tracked_actor)

        return started_actors

    def _enqueue_cached_actor_tasks(self, tracked_actor: TrackedActor):
        assert tracked_actor in self._live_actors_to_ray_actors_resources

        # Enqueue cached futures
        cached_tasks = self._pending_actors_to_enqueued_actor_tasks.pop(
            tracked_actor, []
        )
        for tracked_actor_task, method_name, args, kwargs in cached_tasks:
            self._schedule_tracked_actor_task(
                tracked_actor_task=tracked_actor_task,
                method_name=method_name,
                args=args,
                kwargs=kwargs,
            )

    def _try_kill_actor(self) -> bool:
        """Try to kill actor scheduled for termination."""
        if not self._live_actors_to_kill:
            return False

        tracked_actor = self._live_actors_to_kill.pop()

        # Remove from tracked actors
        (
            ray_actor,
            acquired_resources,
        ) = self._live_actors_to_ray_actors_resources[tracked_actor]

        # Hard kill if requested
        ray.kill(ray_actor)

        self._cleanup_actor_futures(tracked_actor)

        self._actor_stop_resolved(tracked_actor)

        return True

    def _cleanup_actor(self, tracked_actor: TrackedActor):
        self._cleanup_actor_futures(tracked_actor)

        # Remove from tracked actors
        (
            ray_actor,
            acquired_resources,
        ) = self._live_actors_to_ray_actors_resources.pop(tracked_actor)
        self._live_resource_cache = None

        # Return resources
        self._resource_manager.free_resources(acquired_resource=acquired_resources)

    @property
    def all_actors(self) -> List[TrackedActor]:
        """Return all ``TrackedActor`` objects managed by this manager instance."""
        return self.live_actors + self.pending_actors

    @property
    def live_actors(self) -> List[TrackedActor]:
        """Return all ``TrackedActor`` objects that are currently alive."""
        return list(self._live_actors_to_ray_actors_resources)

    @property
    def pending_actors(self) -> List[TrackedActor]:
        """Return all ``TrackedActor`` objects that are currently pending."""
        return list(self._pending_actors_to_attrs)

    @property
    def num_live_actors(self):
        """Return number of started actors."""
        return len(self.live_actors)

    @property
    def num_pending_actors(self) -> int:
        """Return number of pending (not yet started) actors."""
        return len(self.pending_actors)

    @property
    def num_total_actors(self):
        """Return number of total actors."""
        return len(self.all_actors)

    @property
    def num_actor_tasks(self):
        """Return number of pending tasks"""
        return self._actor_task_events.num_futures

    def get_live_actors_resources(self):
        if self._live_resource_cache:
            return self._live_resource_cache

        counter = Counter()
        for _, acq in self._live_actors_to_ray_actors_resources.values():
            for bdl in acq.resource_request.bundles:
                counter.update(bdl)
        self._live_resource_cache = dict(counter)
        return self._live_resource_cache

    def add_actor(
        self,
        cls: Union[Type, ray.actor.ActorClass],
        kwargs: Dict[str, Any],
        resource_request: ResourceRequest,
        *,
        on_start: Optional[Callable[[TrackedActor], None]] = None,
        on_stop: Optional[Callable[[TrackedActor], None]] = None,
        on_error: Optional[Callable[[TrackedActor, Exception], None]] = None,
    ) -> TrackedActor:
        """Add an actor to be tracked.

        This method will request resources to start the actor. Once the resources
        are available, the actor will be started and the
        :meth:`TrackedActor.on_start
        <ray.air.execution._internal.tracked_actor.TrackedActor.on_start>` callback
        will be invoked.

        Args:
            cls: Actor class to schedule.
            kwargs: Keyword arguments to pass to actor class on construction.
            resource_request: Resources required to start the actor.
            on_start: Callback to invoke when the actor started.
            on_stop: Callback to invoke when the actor stopped.
            on_error: Callback to invoke when the actor failed.

        Returns:
            Tracked actor object to reference actor in subsequent API calls.

        """
        tracked_actor = TrackedActor(
            uuid.uuid4().int, on_start=on_start, on_stop=on_stop, on_error=on_error
        )

        self._pending_actors_to_attrs[tracked_actor] = cls, kwargs, resource_request
        self._resource_request_to_pending_actors[resource_request].append(tracked_actor)

        self._resource_manager.request_resources(resource_request=resource_request)

        return tracked_actor

    def remove_actor(
        self,
        tracked_actor: TrackedActor,
        kill: bool = False,
        stop_future: Optional[ray.ObjectRef] = None,
    ) -> bool:
        """Remove a tracked actor.

        If the actor has already been started, this will stop the actor. This will
        trigger the :meth:`TrackedActor.on_stop
        <ray.air.execution._internal.tracked_actor.TrackedActor.on_stop>`
        callback once the actor stopped.

        If the actor has only been requested, but not started, yet, this will cancel
        the actor request. This will not trigger any callback.

        If ``kill=True``, this will use ``ray.kill()`` to forcefully terminate the
        actor. Otherwise, graceful actor deconstruction will be scheduled after
        all currently tracked futures are resolved.

        This method returns a boolean, indicating if a stop future is tracked and
        the ``on_stop`` callback will be invoked. If the actor has been alive,
        this will be ``True``. If the actor hasn't been scheduled, yet, or failed
        (and triggered the ``on_error`` callback), this will be ``False``.

        Args:
            tracked_actor: Tracked actor to be removed.
            kill: If set, will forcefully terminate the actor instead of gracefully
                scheduling termination.
            stop_future: If set, use this future to track actor termination.
                Otherwise, schedule a ``__ray_terminate__`` future.

        Returns:
            Boolean indicating if the actor was previously alive, and thus whether
            a callback will be invoked once it is terminated.

        """
        if tracked_actor.actor_id in self._failed_actor_ids:
            logger.debug(
                f"Tracked actor already failed, no need to remove: {tracked_actor}"
            )
            return False
        elif tracked_actor in self._live_actors_to_ray_actors_resources:
            # Ray actor is running.

            if not kill:
                # Schedule __ray_terminate__ future
                ray_actor, _ = self._live_actors_to_ray_actors_resources[tracked_actor]

                # Clear state futures here to avoid resolving __ray_ready__ futures
                for future in list(
                    self._tracked_actors_to_state_futures[tracked_actor]
                ):
                    self._actor_state_events.discard_future(future)
                    self._tracked_actors_to_state_futures[tracked_actor].remove(future)

                    # If the __ray_ready__ future hasn't resolved yet, but we already
                    # scheduled the actor via Actor.remote(), we just want to stop
                    # it but not trigger any callbacks. This is in accordance with
                    # the contract defined in the docstring.
                    tracked_actor._on_start = None
                    tracked_actor._on_stop = None
                    tracked_actor._on_error = None

                def on_actor_stop(*args, **kwargs):
                    self._actor_stop_resolved(tracked_actor=tracked_actor)

                if stop_future:
                    # If the stop future was schedule via the actor manager,
                    # discard (track it as state future instead).
                    self._actor_task_events.discard_future(stop_future)
                else:
                    stop_future = ray_actor.__ray_terminate__.remote()

                self._actor_state_events.track_future(
                    future=stop_future,
                    on_result=on_actor_stop,
                    on_error=on_actor_stop,
                )

                self._tracked_actors_to_state_futures[tracked_actor].add(stop_future)
            else:
                # kill = True
                self._live_actors_to_kill.add(tracked_actor)

            return True

        elif tracked_actor in self._pending_actors_to_attrs:
            # Actor is pending, stop
            _, _, resource_request = self._pending_actors_to_attrs.pop(tracked_actor)
            self._resource_request_to_pending_actors[resource_request].remove(
                tracked_actor
            )
            self._resource_manager.cancel_resource_request(
                resource_request=resource_request
            )
            return False
        else:
            raise ValueError(f"Unknown tracked actor: {tracked_actor}")

    def is_actor_started(self, tracked_actor: TrackedActor) -> bool:
        """Returns True if the actor has been started.

        Args:
            tracked_actor: Tracked actor object.
        """
        return (
            tracked_actor in self._live_actors_to_ray_actors_resources
            and tracked_actor.actor_id not in self._failed_actor_ids
        )

    def is_actor_failed(self, tracked_actor: TrackedActor) -> bool:
        return tracked_actor.actor_id in self._failed_actor_ids

    def get_actor_resources(
        self, tracked_actor: TrackedActor
    ) -> Optional[AcquiredResources]:
        """Returns the acquired resources of an actor that has been started.

        This will return ``None`` if the actor has not been started, yet.

        Args:
            tracked_actor: Tracked actor object.
        """
        if not self.is_actor_started(tracked_actor):
            return None

        return self._live_actors_to_ray_actors_resources[tracked_actor][1]

    def schedule_actor_task(
        self,
        tracked_actor: TrackedActor,
        method_name: str,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict] = None,
        on_result: Optional[Callable[[TrackedActor, Any], None]] = None,
        on_error: Optional[Callable[[TrackedActor, Exception], None]] = None,
        _return_future: bool = False,
    ) -> Optional[ray.ObjectRef]:
        """Schedule and track a task on an actor.

        This method will schedule a remote task ``method_name`` on the
        ``tracked_actor``.

        This method accepts two optional callbacks that will be invoked when
        their respective events are triggered.

        The ``on_result`` callback is triggered when a task resolves successfully.
        It should accept two arguments: The actor for which the
        task resolved, and the result received from the remote call.

        The ``on_error`` callback is triggered when a task fails.
        It should accept two arguments: The actor for which the
        task threw an error, and the exception.

        Args:
            tracked_actor: Actor to schedule task on.
            method_name: Remote method name to invoke on the actor. If this is
                e.g. ``foo``, then ``actor.foo.remote(*args, **kwargs)`` will be
                scheduled.
            args: Arguments to pass to the task.
            kwargs: Keyword arguments to pass to the task.
            on_result: Callback to invoke when the task resolves.
            on_error: Callback to invoke when the task fails.

        Raises:
            ValueError: If the ``tracked_actor`` is not managed by this event manager.

        """
        args = args or tuple()
        kwargs = kwargs or {}

        if tracked_actor.actor_id in self._failed_actor_ids:
            return

        tracked_actor_task = TrackedActorTask(
            tracked_actor=tracked_actor, on_result=on_result, on_error=on_error
        )

        if tracked_actor not in self._live_actors_to_ray_actors_resources:
            # Actor is not started, yet
            if tracked_actor not in self._pending_actors_to_attrs:
                raise ValueError(
                    f"Tracked actor is not managed by this event manager: "
                    f"{tracked_actor}"
                )

            # Cache tasks for future execution
            self._pending_actors_to_enqueued_actor_tasks[tracked_actor].append(
                (tracked_actor_task, method_name, args, kwargs)
            )
        else:
            res = self._schedule_tracked_actor_task(
                tracked_actor_task=tracked_actor_task,
                method_name=method_name,
                args=args,
                kwargs=kwargs,
                _return_future=_return_future,
            )
            if _return_future:
                return res[1]

    def _schedule_tracked_actor_task(
        self,
        tracked_actor_task: TrackedActorTask,
        method_name: str,
        *,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict] = None,
        _return_future: bool = False,
    ) -> Union[TrackedActorTask, Tuple[TrackedActorTask, ray.ObjectRef]]:
        tracked_actor = tracked_actor_task._tracked_actor
        ray_actor, _ = self._live_actors_to_ray_actors_resources[tracked_actor]

        try:
            remote_fn = getattr(ray_actor, method_name)
        except AttributeError as e:
            raise AttributeError(
                f"Remote function `{method_name}()` does not exist for this actor."
            ) from e

        def on_result(result: Any):
            self._actor_task_resolved(
                tracked_actor_task=tracked_actor_task, result=result
            )

        def on_error(exception: Exception):
            self._actor_task_failed(
                tracked_actor_task=tracked_actor_task, exception=exception
            )

        future = remote_fn.remote(*args, **kwargs)

        self._actor_task_events.track_future(
            future=future, on_result=on_result, on_error=on_error
        )

        self._tracked_actors_to_task_futures[tracked_actor].add(future)

        if _return_future:
            return tracked_actor_task, future

        return tracked_actor_task

    def schedule_actor_tasks(
        self,
        tracked_actors: List[TrackedActor],
        method_name: str,
        *,
        args: Optional[Union[Tuple, List[Tuple]]] = None,
        kwargs: Optional[Union[Dict, List[Dict]]] = None,
        on_result: Optional[Callable[[TrackedActor, Any], None]] = None,
        on_error: Optional[Callable[[TrackedActor, Exception], None]] = None,
    ) -> None:
        """Schedule and track tasks on a list of actors.

        This method will schedule a remote task ``method_name`` on all
        ``tracked_actors``.

        ``args`` and ``kwargs`` can be a single tuple/dict, in which case the same
        (keyword) arguments are passed to all actors. If a list is passed instead,
        they are mapped to the respective actors. In that case, the list of
        (keyword) arguments must be the same length as the list of actors.

        This method accepts two optional callbacks that will be invoked when
        their respective events are triggered.

        The ``on_result`` callback is triggered when a task resolves successfully.
        It should accept two arguments: The actor for which the
        task resolved, and the result received from the remote call.

        The ``on_error`` callback is triggered when a task fails.
        It should accept two arguments: The actor for which the
        task threw an error, and the exception.

        Args:
            tracked_actors: List of actors to schedule tasks on.
            method_name: Remote actor method to invoke on the actors. If this is
                e.g. ``foo``, then ``actor.foo.remote(*args, **kwargs)`` will be
                scheduled on all actors.
            args: Arguments to pass to the task.
            kwargs: Keyword arguments to pass to the task.
            on_result: Callback to invoke when the task resolves.
            on_error: Callback to invoke when the task fails.

        """
        if not isinstance(args, List):
            args_list = [args] * len(tracked_actors)
        else:
            if len(tracked_actors) != len(args):
                raise ValueError(
                    f"Length of args must be the same as tracked_actors "
                    f"list. Got `len(kwargs)={len(kwargs)}` and "
                    f"`len(tracked_actors)={len(tracked_actors)}"
                )
            args_list = args

        if not isinstance(kwargs, List):
            kwargs_list = [kwargs] * len(tracked_actors)
        else:
            if len(tracked_actors) != len(kwargs):
                raise ValueError(
                    f"Length of kwargs must be the same as tracked_actors "
                    f"list. Got `len(args)={len(args)}` and "
                    f"`len(tracked_actors)={len(tracked_actors)}"
                )
            kwargs_list = kwargs

        for tracked_actor, args, kwargs in zip(tracked_actors, args_list, kwargs_list):
            self.schedule_actor_task(
                tracked_actor=tracked_actor,
                method_name=method_name,
                args=args,
                kwargs=kwargs,
                on_result=on_result,
                on_error=on_error,
            )

    def clear_actor_task_futures(self, tracked_actor: TrackedActor):
        """Discard all actor task futures from a tracked actor."""
        futures = self._tracked_actors_to_task_futures.pop(tracked_actor, [])
        for future in futures:
            self._actor_task_events.discard_future(future)

    def _cleanup_actor_futures(self, tracked_actor: TrackedActor):
        # Remove all actor task futures
        self.clear_actor_task_futures(tracked_actor=tracked_actor)

        # Remove all actor state futures
        futures = self._tracked_actors_to_state_futures.pop(tracked_actor, [])
        for future in futures:
            self._actor_state_events.discard_future(future)

    def cleanup(self):
        for (
            actor,
            acquired_resources,
        ) in self._live_actors_to_ray_actors_resources.values():
            ray.kill(actor)
            self._resource_manager.free_resources(acquired_resources)

        for (
            resource_request,
            pending_actors,
        ) in self._resource_request_to_pending_actors.items():
            for i in range(len(pending_actors)):
                self._resource_manager.cancel_resource_request(resource_request)

        self._resource_manager.clear()

        self.__init__(resource_manager=self._resource_manager)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/node_manager.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as src_dot_ray_dot_protobuf_dot_common__pb2
from . import gcs_pb2 as src_dot_ray_dot_protobuf_dot_gcs__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#src/ray/protobuf/node_manager.proto\x12\x07ray.rpc\x1a\x1dsrc/ray/protobuf/common.proto\x1a\x1asrc/ray/protobuf/gcs.proto\"p\n\x13WorkerBacklogReport\x12\x36\n\rresource_spec\x18\x01 \x01(\x0b\x32\x11.ray.rpc.TaskSpecR\x0cresourceSpec\x12!\n\x0c\x62\x61\x63klog_size\x18\x02 \x01(\x03R\x0b\x62\x61\x63klogSize\"\x80\x01\n\x1aReportWorkerBacklogRequest\x12\x1b\n\tworker_id\x18\x01 \x01(\x0cR\x08workerId\x12\x45\n\x0f\x62\x61\x63klog_reports\x18\x02 \x03(\x0b\x32\x1c.ray.rpc.WorkerBacklogReportR\x0e\x62\x61\x63klogReports\"\x1a\n\x18ReportWorkerBacklogReply\"\xe0\x01\n\x19RequestWorkerLeaseRequest\x12\x36\n\rresource_spec\x18\x01 \x01(\x0b\x32\x11.ray.rpc.TaskSpecR\x0cresourceSpec\x12!\n\x0c\x62\x61\x63klog_size\x18\x02 \x01(\x03R\x0b\x62\x61\x63klogSize\x12&\n\x0fgrant_or_reject\x18\x03 \x01(\x08R\rgrantOrReject\x12@\n\x1dis_selected_based_on_locality\x18\x04 \x01(\x08R\x19isSelectedBasedOnLocality\"\x81\x06\n\x17RequestWorkerLeaseReply\x12\x37\n\x0eworker_address\x18\x01 \x01(\x0b\x32\x10.ray.rpc.AddressR\rworkerAddress\x12G\n\x17retry_at_raylet_address\x18\x02 \x01(\x0b\x32\x10.ray.rpc.AddressR\x14retryAtRayletAddress\x12\x44\n\x10resource_mapping\x18\x03 \x03(\x0b\x32\x19.ray.rpc.ResourceMapEntryR\x0fresourceMapping\x12\x1a\n\x08\x63\x61nceled\x18\x04 \x01(\x08R\x08\x63\x61nceled\x12\x1d\n\nworker_pid\x18\x06 \x01(\rR\tworkerPid\x12\x1a\n\x08rejected\x18\x07 \x01(\x08R\x08rejected\x12=\n\x0eresources_data\x18\x08 \x01(\x0b\x32\x16.ray.rpc.ResourcesDataR\rresourcesData\x12Y\n\x0c\x66\x61ilure_type\x18\t \x01(\x0e\x32\x36.ray.rpc.RequestWorkerLeaseReply.SchedulingFailureTypeR\x0b\x66\x61ilureType\x12<\n\x1ascheduling_failure_message\x18\n \x01(\tR\x18schedulingFailureMessage\"\xee\x01\n\x15SchedulingFailureType\x12\x0e\n\nNOT_FAILED\x10\x00\x12\x15\n\x11SCHEDULING_FAILED\x10\x01\x12\x30\n,SCHEDULING_CANCELLED_PLACEMENT_GROUP_REMOVED\x10\x02\x12\x31\n-SCHEDULING_CANCELLED_RUNTIME_ENV_SETUP_FAILED\x10\x03\x12!\n\x1dSCHEDULING_CANCELLED_INTENDED\x10\x04\x12&\n\"SCHEDULING_CANCELLED_UNSCHEDULABLE\x10\x05\"S\n\x1dPrepareBundleResourcesRequest\x12\x32\n\x0c\x62undle_specs\x18\x01 \x03(\x0b\x32\x0f.ray.rpc.BundleR\x0b\x62undleSpecs\"7\n\x1bPrepareBundleResourcesReply\x12\x18\n\x07success\x18\x01 \x01(\x08R\x07success\"R\n\x1c\x43ommitBundleResourcesRequest\x12\x32\n\x0c\x62undle_specs\x18\x01 \x03(\x0b\x32\x0f.ray.rpc.BundleR\x0b\x62undleSpecs\"\x1c\n\x1a\x43ommitBundleResourcesReply\"P\n\x1c\x43\x61ncelResourceReserveRequest\x12\x30\n\x0b\x62undle_spec\x18\x01 \x01(\x0b\x32\x0f.ray.rpc.BundleR\nbundleSpec\"\x1c\n\x1a\x43\x61ncelResourceReserveReply\"\xa7\x01\n\x13ReturnWorkerRequest\x12\x1f\n\x0bworker_port\x18\x01 \x01(\x05R\nworkerPort\x12\x1b\n\tworker_id\x18\x02 \x01(\x0cR\x08workerId\x12+\n\x11\x64isconnect_worker\x18\x03 \x01(\x08R\x10\x64isconnectWorker\x12%\n\x0eworker_exiting\x18\x04 \x01(\x08R\rworkerExiting\"\x13\n\x11ReturnWorkerReply\"H\n\x1bReleaseUnusedWorkersRequest\x12)\n\x11worker_ids_in_use\x18\x01 \x03(\x0cR\x0eworkerIdsInUse\"\x1b\n\x19ReleaseUnusedWorkersReply\"3\n\x15ShutdownRayletRequest\x12\x1a\n\x08graceful\x18\x01 \x01(\x08R\x08graceful\"\x15\n\x13ShutdownRayletReply\"3\n\x18\x43\x61ncelWorkerLeaseRequest\x12\x17\n\x07task_id\x18\x01 \x01(\x0cR\x06taskId\"2\n\x16\x43\x61ncelWorkerLeaseReply\x12\x18\n\x07success\x18\x01 \x01(\x08R\x07success\"\xa4\x01\n\x13PinObjectIDsRequest\x12\x35\n\rowner_address\x18\x01 \x01(\x0b\x32\x10.ray.rpc.AddressR\x0cownerAddress\x12\x1d\n\nobject_ids\x18\x02 \x03(\x0cR\tobjectIds\x12&\n\x0cgenerator_id\x18\x03 \x01(\x0cH\x00R\x0bgeneratorId\x88\x01\x01\x42\x0f\n\r_generator_id\"1\n\x11PinObjectIDsReply\x12\x1c\n\tsuccesses\x18\x01 \x03(\x08R\tsuccesses\"E\n\x13GetNodeStatsRequest\x12.\n\x13include_memory_info\x18\x01 \x01(\x08R\x11includeMemoryInfo\"\xfa\x05\n\x10ObjectStoreStats\x12+\n\x12spill_time_total_s\x18\x01 \x01(\x01R\x0fspillTimeTotalS\x12.\n\x13spilled_bytes_total\x18\x02 \x01(\x03R\x11spilledBytesTotal\x12\x32\n\x15spilled_objects_total\x18\x03 \x01(\x03R\x13spilledObjectsTotal\x12/\n\x14restore_time_total_s\x18\x04 \x01(\x01R\x11restoreTimeTotalS\x12\x30\n\x14restored_bytes_total\x18\x05 \x01(\x03R\x12restoredBytesTotal\x12\x34\n\x16restored_objects_total\x18\x06 \x01(\x03R\x14restoredObjectsTotal\x12\x35\n\x17object_store_bytes_used\x18\x07 \x01(\x03R\x14objectStoreBytesUsed\x12\x37\n\x18object_store_bytes_avail\x18\x08 \x01(\x03R\x15objectStoreBytesAvail\x12\x44\n\x1fobject_store_bytes_primary_copy\x18\t \x01(\x03R\x1bobjectStoreBytesPrimaryCopy\x12=\n\x1bobject_store_bytes_fallback\x18\n \x01(\x03R\x18objectStoreBytesFallback\x12*\n\x11num_local_objects\x18\x0b \x01(\x03R\x0fnumLocalObjects\x12%\n\x0e\x63onsumed_bytes\x18\x0c \x01(\x03R\rconsumedBytes\x12.\n\x13object_pulls_queued\x18\r \x01(\x08R\x11objectPullsQueued\x12\x44\n\x1fnum_object_store_primary_copies\x18\x0e \x01(\x03R\x1bnumObjectStorePrimaryCopies\"\xb8\x01\n\x11GetNodeStatsReply\x12\x46\n\x12\x63ore_workers_stats\x18\x01 \x03(\x0b\x32\x18.ray.rpc.CoreWorkerStatsR\x10\x63oreWorkersStats\x12\x1f\n\x0bnum_workers\x18\x03 \x01(\rR\nnumWorkers\x12:\n\x0bstore_stats\x18\x06 \x01(\x0b\x32\x19.ray.rpc.ObjectStoreStatsR\nstoreStats\"\x11\n\x0fGlobalGCRequest\"\x0f\n\rGlobalGCReply\"O\n\x1d\x46ormatGlobalMemoryInfoRequest\x12.\n\x13include_memory_info\x18\x01 \x01(\x08R\x11includeMemoryInfo\"\x80\x01\n\x1b\x46ormatGlobalMemoryInfoReply\x12%\n\x0ememory_summary\x18\x01 \x01(\tR\rmemorySummary\x12:\n\x0bstore_stats\x18\x02 \x01(\x0b\x32\x19.ray.rpc.ObjectStoreStatsR\nstoreStats\";\n\x1cRequestObjectSpillageRequest\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\"}\n\x1aRequestObjectSpillageReply\x12\x18\n\x07success\x18\x01 \x01(\x08R\x07success\x12\x1d\n\nobject_url\x18\x02 \x01(\tR\tobjectUrl\x12&\n\x0fspilled_node_id\x18\x03 \x01(\x0cR\rspilledNodeId\"T\n\x1bReleaseUnusedBundlesRequest\x12\x35\n\x0e\x62undles_in_use\x18\x01 \x03(\x0b\x32\x0f.ray.rpc.BundleR\x0c\x62undlesInUse\"\x1b\n\x19ReleaseUnusedBundlesReply\"\x18\n\x16GetSystemConfigRequest\";\n\x14GetSystemConfigReply\x12#\n\rsystem_config\x18\x01 \x01(\tR\x0csystemConfig\"\x1e\n\x1cRequestResourceReportRequest\"R\n\x1aRequestResourceReportReply\x12\x34\n\tresources\x18\x01 \x01(\x0b\x32\x16.ray.rpc.ResourcesDataR\tresources\"c\n\x1aUpdateResourceUsageRequest\x12\x45\n\x1fserialized_resource_usage_batch\x18\x01 \x01(\x0cR\x1cserializedResourceUsageBatch\"\x1a\n\x18UpdateResourceUsageReply\":\n\x13GetTasksInfoRequest\x12\x19\n\x05limit\x18\x01 \x01(\x03H\x00R\x05limit\x88\x01\x01\x42\x08\n\x06_limit\"\xa2\x01\n\x11GetTasksInfoReply\x12M\n\x17owned_task_info_entries\x18\x01 \x03(\x0b\x32\x16.ray.rpc.TaskInfoEntryR\x14ownedTaskInfoEntries\x12(\n\x10running_task_ids\x18\x02 \x03(\x0cR\x0erunningTaskIds\x12\x14\n\x05total\x18\x03 \x01(\x03R\x05total\"<\n\x15GetObjectsInfoRequest\x12\x19\n\x05limit\x18\x01 \x01(\x03H\x00R\x05limit\x88\x01\x01\x42\x08\n\x06_limit\"s\n\x13GetObjectsInfoReply\x12\x46\n\x12\x63ore_workers_stats\x18\x01 \x03(\x0b\x32\x18.ray.rpc.CoreWorkerStatsR\x10\x63oreWorkersStats\x12\x14\n\x05total\x18\x02 \x01(\x03R\x05total\"\x18\n\x16GetResourceLoadRequest\"L\n\x14GetResourceLoadReply\x12\x34\n\tresources\x18\x01 \x01(\x0b\x32\x16.ray.rpc.ResourcesDataR\tresources\"\x19\n\x17NotifyGCSRestartRequest\"\x17\n\x15NotifyGCSRestartReply\"5\n\x1aGetTaskFailureCauseRequest\x12\x17\n\x07task_id\x18\x01 \x01(\x0cR\x06taskId\"\xa1\x01\n\x18GetTaskFailureCauseReply\x12?\n\rfailure_cause\x18\x01 \x01(\x0b\x32\x15.ray.rpc.RayErrorInfoH\x00R\x0c\x66\x61ilureCause\x88\x01\x01\x12\x32\n\x15\x66\x61il_task_immediately\x18\x02 \x01(\x08R\x13\x66\x61ilTaskImmediatelyB\x10\n\x0e_failure_cause2\x90\x10\n\x12NodeManagerService\x12]\n\x13UpdateResourceUsage\x12#.ray.rpc.UpdateResourceUsageRequest\x1a!.ray.rpc.UpdateResourceUsageReply\x12T\n\x10NotifyGCSRestart\x12 .ray.rpc.NotifyGCSRestartRequest\x1a\x1e.ray.rpc.NotifyGCSRestartReply\x12Q\n\x0fGetResourceLoad\x12\x1f.ray.rpc.GetResourceLoadRequest\x1a\x1d.ray.rpc.GetResourceLoadReply\x12\x63\n\x15RequestResourceReport\x12%.ray.rpc.RequestResourceReportRequest\x1a#.ray.rpc.RequestResourceReportReply\x12Z\n\x12RequestWorkerLease\x12\".ray.rpc.RequestWorkerLeaseRequest\x1a .ray.rpc.RequestWorkerLeaseReply\x12]\n\x13ReportWorkerBacklog\x12#.ray.rpc.ReportWorkerBacklogRequest\x1a!.ray.rpc.ReportWorkerBacklogReply\x12H\n\x0cReturnWorker\x12\x1c.ray.rpc.ReturnWorkerRequest\x1a\x1a.ray.rpc.ReturnWorkerReply\x12`\n\x14ReleaseUnusedWorkers\x12$.ray.rpc.ReleaseUnusedWorkersRequest\x1a\".ray.rpc.ReleaseUnusedWorkersReply\x12N\n\x0eShutdownRaylet\x12\x1e.ray.rpc.ShutdownRayletRequest\x1a\x1c.ray.rpc.ShutdownRayletReply\x12\x66\n\x16PrepareBundleResources\x12&.ray.rpc.PrepareBundleResourcesRequest\x1a$.ray.rpc.PrepareBundleResourcesReply\x12\x63\n\x15\x43ommitBundleResources\x12%.ray.rpc.CommitBundleResourcesRequest\x1a#.ray.rpc.CommitBundleResourcesReply\x12\x63\n\x15\x43\x61ncelResourceReserve\x12%.ray.rpc.CancelResourceReserveRequest\x1a#.ray.rpc.CancelResourceReserveReply\x12W\n\x11\x43\x61ncelWorkerLease\x12!.ray.rpc.CancelWorkerLeaseRequest\x1a\x1f.ray.rpc.CancelWorkerLeaseReply\x12H\n\x0cPinObjectIDs\x12\x1c.ray.rpc.PinObjectIDsRequest\x1a\x1a.ray.rpc.PinObjectIDsReply\x12H\n\x0cGetNodeStats\x12\x1c.ray.rpc.GetNodeStatsRequest\x1a\x1a.ray.rpc.GetNodeStatsReply\x12<\n\x08GlobalGC\x12\x18.ray.rpc.GlobalGCRequest\x1a\x16.ray.rpc.GlobalGCReply\x12\x66\n\x16\x46ormatGlobalMemoryInfo\x12&.ray.rpc.FormatGlobalMemoryInfoRequest\x1a$.ray.rpc.FormatGlobalMemoryInfoReply\x12\x63\n\x15RequestObjectSpillage\x12%.ray.rpc.RequestObjectSpillageRequest\x1a#.ray.rpc.RequestObjectSpillageReply\x12`\n\x14ReleaseUnusedBundles\x12$.ray.rpc.ReleaseUnusedBundlesRequest\x1a\".ray.rpc.ReleaseUnusedBundlesReply\x12Q\n\x0fGetSystemConfig\x12\x1f.ray.rpc.GetSystemConfigRequest\x1a\x1d.ray.rpc.GetSystemConfigReply\x12H\n\x0cGetTasksInfo\x12\x1c.ray.rpc.GetTasksInfoRequest\x1a\x1a.ray.rpc.GetTasksInfoReply\x12N\n\x0eGetObjectsInfo\x12\x1e.ray.rpc.GetObjectsInfoRequest\x1a\x1c.ray.rpc.GetObjectsInfoReply\x12]\n\x13GetTaskFailureCause\x12#.ray.rpc.GetTaskFailureCauseRequest\x1a!.ray.rpc.GetTaskFailureCauseReplyB\x03\xf8\x01\x01\x62\x06proto3')



_WORKERBACKLOGREPORT = DESCRIPTOR.message_types_by_name['WorkerBacklogReport']
_REPORTWORKERBACKLOGREQUEST = DESCRIPTOR.message_types_by_name['ReportWorkerBacklogRequest']
_REPORTWORKERBACKLOGREPLY = DESCRIPTOR.message_types_by_name['ReportWorkerBacklogReply']
_REQUESTWORKERLEASEREQUEST = DESCRIPTOR.message_types_by_name['RequestWorkerLeaseRequest']
_REQUESTWORKERLEASEREPLY = DESCRIPTOR.message_types_by_name['RequestWorkerLeaseReply']
_PREPAREBUNDLERESOURCESREQUEST = DESCRIPTOR.message_types_by_name['PrepareBundleResourcesRequest']
_PREPAREBUNDLERESOURCESREPLY = DESCRIPTOR.message_types_by_name['PrepareBundleResourcesReply']
_COMMITBUNDLERESOURCESREQUEST = DESCRIPTOR.message_types_by_name['CommitBundleResourcesRequest']
_COMMITBUNDLERESOURCESREPLY = DESCRIPTOR.message_types_by_name['CommitBundleResourcesReply']
_CANCELRESOURCERESERVEREQUEST = DESCRIPTOR.message_types_by_name['CancelResourceReserveRequest']
_CANCELRESOURCERESERVEREPLY = DESCRIPTOR.message_types_by_name['CancelResourceReserveReply']
_RETURNWORKERREQUEST = DESCRIPTOR.message_types_by_name['ReturnWorkerRequest']
_RETURNWORKERREPLY = DESCRIPTOR.message_types_by_name['ReturnWorkerReply']
_RELEASEUNUSEDWORKERSREQUEST = DESCRIPTOR.message_types_by_name['ReleaseUnusedWorkersRequest']
_RELEASEUNUSEDWORKERSREPLY = DESCRIPTOR.message_types_by_name['ReleaseUnusedWorkersReply']
_SHUTDOWNRAYLETREQUEST = DESCRIPTOR.message_types_by_name['ShutdownRayletRequest']
_SHUTDOWNRAYLETREPLY = DESCRIPTOR.message_types_by_name['ShutdownRayletReply']
_CANCELWORKERLEASEREQUEST = DESCRIPTOR.message_types_by_name['CancelWorkerLeaseRequest']
_CANCELWORKERLEASEREPLY = DESCRIPTOR.message_types_by_name['CancelWorkerLeaseReply']
_PINOBJECTIDSREQUEST = DESCRIPTOR.message_types_by_name['PinObjectIDsRequest']
_PINOBJECTIDSREPLY = DESCRIPTOR.message_types_by_name['PinObjectIDsReply']
_GETNODESTATSREQUEST = DESCRIPTOR.message_types_by_name['GetNodeStatsRequest']
_OBJECTSTORESTATS = DESCRIPTOR.message_types_by_name['ObjectStoreStats']
_GETNODESTATSREPLY = DESCRIPTOR.message_types_by_name['GetNodeStatsReply']
_GLOBALGCREQUEST = DESCRIPTOR.message_types_by_name['GlobalGCRequest']
_GLOBALGCREPLY = DESCRIPTOR.message_types_by_name['GlobalGCReply']
_FORMATGLOBALMEMORYINFOREQUEST = DESCRIPTOR.message_types_by_name['FormatGlobalMemoryInfoRequest']
_FORMATGLOBALMEMORYINFOREPLY = DESCRIPTOR.message_types_by_name['FormatGlobalMemoryInfoReply']
_REQUESTOBJECTSPILLAGEREQUEST = DESCRIPTOR.message_types_by_name['RequestObjectSpillageRequest']
_REQUESTOBJECTSPILLAGEREPLY = DESCRIPTOR.message_types_by_name['RequestObjectSpillageReply']
_RELEASEUNUSEDBUNDLESREQUEST = DESCRIPTOR.message_types_by_name['ReleaseUnusedBundlesRequest']
_RELEASEUNUSEDBUNDLESREPLY = DESCRIPTOR.message_types_by_name['ReleaseUnusedBundlesReply']
_GETSYSTEMCONFIGREQUEST = DESCRIPTOR.message_types_by_name['GetSystemConfigRequest']
_GETSYSTEMCONFIGREPLY = DESCRIPTOR.message_types_by_name['GetSystemConfigReply']
_REQUESTRESOURCEREPORTREQUEST = DESCRIPTOR.message_types_by_name['RequestResourceReportRequest']
_REQUESTRESOURCEREPORTREPLY = DESCRIPTOR.message_types_by_name['RequestResourceReportReply']
_UPDATERESOURCEUSAGEREQUEST = DESCRIPTOR.message_types_by_name['UpdateResourceUsageRequest']
_UPDATERESOURCEUSAGEREPLY = DESCRIPTOR.message_types_by_name['UpdateResourceUsageReply']
_GETTASKSINFOREQUEST = DESCRIPTOR.message_types_by_name['GetTasksInfoRequest']
_GETTASKSINFOREPLY = DESCRIPTOR.message_types_by_name['GetTasksInfoReply']
_GETOBJECTSINFOREQUEST = DESCRIPTOR.message_types_by_name['GetObjectsInfoRequest']
_GETOBJECTSINFOREPLY = DESCRIPTOR.message_types_by_name['GetObjectsInfoReply']
_GETRESOURCELOADREQUEST = DESCRIPTOR.message_types_by_name['GetResourceLoadRequest']
_GETRESOURCELOADREPLY = DESCRIPTOR.message_types_by_name['GetResourceLoadReply']
_NOTIFYGCSRESTARTREQUEST = DESCRIPTOR.message_types_by_name['NotifyGCSRestartRequest']
_NOTIFYGCSRESTARTREPLY = DESCRIPTOR.message_types_by_name['NotifyGCSRestartReply']
_GETTASKFAILURECAUSEREQUEST = DESCRIPTOR.message_types_by_name['GetTaskFailureCauseRequest']
_GETTASKFAILURECAUSEREPLY = DESCRIPTOR.message_types_by_name['GetTaskFailureCauseReply']
_REQUESTWORKERLEASEREPLY_SCHEDULINGFAILURETYPE = _REQUESTWORKERLEASEREPLY.enum_types_by_name['SchedulingFailureType']
WorkerBacklogReport = _reflection.GeneratedProtocolMessageType('WorkerBacklogReport', (_message.Message,), {
  'DESCRIPTOR' : _WORKERBACKLOGREPORT,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerBacklogReport)
  })
_sym_db.RegisterMessage(WorkerBacklogReport)

ReportWorkerBacklogRequest = _reflection.GeneratedProtocolMessageType('ReportWorkerBacklogRequest', (_message.Message,), {
  'DESCRIPTOR' : _REPORTWORKERBACKLOGREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReportWorkerBacklogRequest)
  })
_sym_db.RegisterMessage(ReportWorkerBacklogRequest)

ReportWorkerBacklogReply = _reflection.GeneratedProtocolMessageType('ReportWorkerBacklogReply', (_message.Message,), {
  'DESCRIPTOR' : _REPORTWORKERBACKLOGREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReportWorkerBacklogReply)
  })
_sym_db.RegisterMessage(ReportWorkerBacklogReply)

RequestWorkerLeaseRequest = _reflection.GeneratedProtocolMessageType('RequestWorkerLeaseRequest', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTWORKERLEASEREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RequestWorkerLeaseRequest)
  })
_sym_db.RegisterMessage(RequestWorkerLeaseRequest)

RequestWorkerLeaseReply = _reflection.GeneratedProtocolMessageType('RequestWorkerLeaseReply', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTWORKERLEASEREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RequestWorkerLeaseReply)
  })
_sym_db.RegisterMessage(RequestWorkerLeaseReply)

PrepareBundleResourcesRequest = _reflection.GeneratedProtocolMessageType('PrepareBundleResourcesRequest', (_message.Message,), {
  'DESCRIPTOR' : _PREPAREBUNDLERESOURCESREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PrepareBundleResourcesRequest)
  })
_sym_db.RegisterMessage(PrepareBundleResourcesRequest)

PrepareBundleResourcesReply = _reflection.GeneratedProtocolMessageType('PrepareBundleResourcesReply', (_message.Message,), {
  'DESCRIPTOR' : _PREPAREBUNDLERESOURCESREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PrepareBundleResourcesReply)
  })
_sym_db.RegisterMessage(PrepareBundleResourcesReply)

CommitBundleResourcesRequest = _reflection.GeneratedProtocolMessageType('CommitBundleResourcesRequest', (_message.Message,), {
  'DESCRIPTOR' : _COMMITBUNDLERESOURCESREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CommitBundleResourcesRequest)
  })
_sym_db.RegisterMessage(CommitBundleResourcesRequest)

CommitBundleResourcesReply = _reflection.GeneratedProtocolMessageType('CommitBundleResourcesReply', (_message.Message,), {
  'DESCRIPTOR' : _COMMITBUNDLERESOURCESREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CommitBundleResourcesReply)
  })
_sym_db.RegisterMessage(CommitBundleResourcesReply)

CancelResourceReserveRequest = _reflection.GeneratedProtocolMessageType('CancelResourceReserveRequest', (_message.Message,), {
  'DESCRIPTOR' : _CANCELRESOURCERESERVEREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CancelResourceReserveRequest)
  })
_sym_db.RegisterMessage(CancelResourceReserveRequest)

CancelResourceReserveReply = _reflection.GeneratedProtocolMessageType('CancelResourceReserveReply', (_message.Message,), {
  'DESCRIPTOR' : _CANCELRESOURCERESERVEREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CancelResourceReserveReply)
  })
_sym_db.RegisterMessage(CancelResourceReserveReply)

ReturnWorkerRequest = _reflection.GeneratedProtocolMessageType('ReturnWorkerRequest', (_message.Message,), {
  'DESCRIPTOR' : _RETURNWORKERREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReturnWorkerRequest)
  })
_sym_db.RegisterMessage(ReturnWorkerRequest)

ReturnWorkerReply = _reflection.GeneratedProtocolMessageType('ReturnWorkerReply', (_message.Message,), {
  'DESCRIPTOR' : _RETURNWORKERREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReturnWorkerReply)
  })
_sym_db.RegisterMessage(ReturnWorkerReply)

ReleaseUnusedWorkersRequest = _reflection.GeneratedProtocolMessageType('ReleaseUnusedWorkersRequest', (_message.Message,), {
  'DESCRIPTOR' : _RELEASEUNUSEDWORKERSREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReleaseUnusedWorkersRequest)
  })
_sym_db.RegisterMessage(ReleaseUnusedWorkersRequest)

ReleaseUnusedWorkersReply = _reflection.GeneratedProtocolMessageType('ReleaseUnusedWorkersReply', (_message.Message,), {
  'DESCRIPTOR' : _RELEASEUNUSEDWORKERSREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReleaseUnusedWorkersReply)
  })
_sym_db.RegisterMessage(ReleaseUnusedWorkersReply)

ShutdownRayletRequest = _reflection.GeneratedProtocolMessageType('ShutdownRayletRequest', (_message.Message,), {
  'DESCRIPTOR' : _SHUTDOWNRAYLETREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ShutdownRayletRequest)
  })
_sym_db.RegisterMessage(ShutdownRayletRequest)

ShutdownRayletReply = _reflection.GeneratedProtocolMessageType('ShutdownRayletReply', (_message.Message,), {
  'DESCRIPTOR' : _SHUTDOWNRAYLETREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ShutdownRayletReply)
  })
_sym_db.RegisterMessage(ShutdownRayletReply)

CancelWorkerLeaseRequest = _reflection.GeneratedProtocolMessageType('CancelWorkerLeaseRequest', (_message.Message,), {
  'DESCRIPTOR' : _CANCELWORKERLEASEREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CancelWorkerLeaseRequest)
  })
_sym_db.RegisterMessage(CancelWorkerLeaseRequest)

CancelWorkerLeaseReply = _reflection.GeneratedProtocolMessageType('CancelWorkerLeaseReply', (_message.Message,), {
  'DESCRIPTOR' : _CANCELWORKERLEASEREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CancelWorkerLeaseReply)
  })
_sym_db.RegisterMessage(CancelWorkerLeaseReply)

PinObjectIDsRequest = _reflection.GeneratedProtocolMessageType('PinObjectIDsRequest', (_message.Message,), {
  'DESCRIPTOR' : _PINOBJECTIDSREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PinObjectIDsRequest)
  })
_sym_db.RegisterMessage(PinObjectIDsRequest)

PinObjectIDsReply = _reflection.GeneratedProtocolMessageType('PinObjectIDsReply', (_message.Message,), {
  'DESCRIPTOR' : _PINOBJECTIDSREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PinObjectIDsReply)
  })
_sym_db.RegisterMessage(PinObjectIDsReply)

GetNodeStatsRequest = _reflection.GeneratedProtocolMessageType('GetNodeStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETNODESTATSREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetNodeStatsRequest)
  })
_sym_db.RegisterMessage(GetNodeStatsRequest)

ObjectStoreStats = _reflection.GeneratedProtocolMessageType('ObjectStoreStats', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTSTORESTATS,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectStoreStats)
  })
_sym_db.RegisterMessage(ObjectStoreStats)

GetNodeStatsReply = _reflection.GeneratedProtocolMessageType('GetNodeStatsReply', (_message.Message,), {
  'DESCRIPTOR' : _GETNODESTATSREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetNodeStatsReply)
  })
_sym_db.RegisterMessage(GetNodeStatsReply)

GlobalGCRequest = _reflection.GeneratedProtocolMessageType('GlobalGCRequest', (_message.Message,), {
  'DESCRIPTOR' : _GLOBALGCREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GlobalGCRequest)
  })
_sym_db.RegisterMessage(GlobalGCRequest)

GlobalGCReply = _reflection.GeneratedProtocolMessageType('GlobalGCReply', (_message.Message,), {
  'DESCRIPTOR' : _GLOBALGCREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GlobalGCReply)
  })
_sym_db.RegisterMessage(GlobalGCReply)

FormatGlobalMemoryInfoRequest = _reflection.GeneratedProtocolMessageType('FormatGlobalMemoryInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _FORMATGLOBALMEMORYINFOREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.FormatGlobalMemoryInfoRequest)
  })
_sym_db.RegisterMessage(FormatGlobalMemoryInfoRequest)

FormatGlobalMemoryInfoReply = _reflection.GeneratedProtocolMessageType('FormatGlobalMemoryInfoReply', (_message.Message,), {
  'DESCRIPTOR' : _FORMATGLOBALMEMORYINFOREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.FormatGlobalMemoryInfoReply)
  })
_sym_db.RegisterMessage(FormatGlobalMemoryInfoReply)

RequestObjectSpillageRequest = _reflection.GeneratedProtocolMessageType('RequestObjectSpillageRequest', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTOBJECTSPILLAGEREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RequestObjectSpillageRequest)
  })
_sym_db.RegisterMessage(RequestObjectSpillageRequest)

RequestObjectSpillageReply = _reflection.GeneratedProtocolMessageType('RequestObjectSpillageReply', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTOBJECTSPILLAGEREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RequestObjectSpillageReply)
  })
_sym_db.RegisterMessage(RequestObjectSpillageReply)

ReleaseUnusedBundlesRequest = _reflection.GeneratedProtocolMessageType('ReleaseUnusedBundlesRequest', (_message.Message,), {
  'DESCRIPTOR' : _RELEASEUNUSEDBUNDLESREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReleaseUnusedBundlesRequest)
  })
_sym_db.RegisterMessage(ReleaseUnusedBundlesRequest)

ReleaseUnusedBundlesReply = _reflection.GeneratedProtocolMessageType('ReleaseUnusedBundlesReply', (_message.Message,), {
  'DESCRIPTOR' : _RELEASEUNUSEDBUNDLESREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ReleaseUnusedBundlesReply)
  })
_sym_db.RegisterMessage(ReleaseUnusedBundlesReply)

GetSystemConfigRequest = _reflection.GeneratedProtocolMessageType('GetSystemConfigRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMCONFIGREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetSystemConfigRequest)
  })
_sym_db.RegisterMessage(GetSystemConfigRequest)

GetSystemConfigReply = _reflection.GeneratedProtocolMessageType('GetSystemConfigReply', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMCONFIGREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetSystemConfigReply)
  })
_sym_db.RegisterMessage(GetSystemConfigReply)

RequestResourceReportRequest = _reflection.GeneratedProtocolMessageType('RequestResourceReportRequest', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTRESOURCEREPORTREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RequestResourceReportRequest)
  })
_sym_db.RegisterMessage(RequestResourceReportRequest)

RequestResourceReportReply = _reflection.GeneratedProtocolMessageType('RequestResourceReportReply', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTRESOURCEREPORTREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RequestResourceReportReply)
  })
_sym_db.RegisterMessage(RequestResourceReportReply)

UpdateResourceUsageRequest = _reflection.GeneratedProtocolMessageType('UpdateResourceUsageRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATERESOURCEUSAGEREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.UpdateResourceUsageRequest)
  })
_sym_db.RegisterMessage(UpdateResourceUsageRequest)

UpdateResourceUsageReply = _reflection.GeneratedProtocolMessageType('UpdateResourceUsageReply', (_message.Message,), {
  'DESCRIPTOR' : _UPDATERESOURCEUSAGEREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.UpdateResourceUsageReply)
  })
_sym_db.RegisterMessage(UpdateResourceUsageReply)

GetTasksInfoRequest = _reflection.GeneratedProtocolMessageType('GetTasksInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKSINFOREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetTasksInfoRequest)
  })
_sym_db.RegisterMessage(GetTasksInfoRequest)

GetTasksInfoReply = _reflection.GeneratedProtocolMessageType('GetTasksInfoReply', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKSINFOREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetTasksInfoReply)
  })
_sym_db.RegisterMessage(GetTasksInfoReply)

GetObjectsInfoRequest = _reflection.GeneratedProtocolMessageType('GetObjectsInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETOBJECTSINFOREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetObjectsInfoRequest)
  })
_sym_db.RegisterMessage(GetObjectsInfoRequest)

GetObjectsInfoReply = _reflection.GeneratedProtocolMessageType('GetObjectsInfoReply', (_message.Message,), {
  'DESCRIPTOR' : _GETOBJECTSINFOREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetObjectsInfoReply)
  })
_sym_db.RegisterMessage(GetObjectsInfoReply)

GetResourceLoadRequest = _reflection.GeneratedProtocolMessageType('GetResourceLoadRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRESOURCELOADREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetResourceLoadRequest)
  })
_sym_db.RegisterMessage(GetResourceLoadRequest)

GetResourceLoadReply = _reflection.GeneratedProtocolMessageType('GetResourceLoadReply', (_message.Message,), {
  'DESCRIPTOR' : _GETRESOURCELOADREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetResourceLoadReply)
  })
_sym_db.RegisterMessage(GetResourceLoadReply)

NotifyGCSRestartRequest = _reflection.GeneratedProtocolMessageType('NotifyGCSRestartRequest', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFYGCSRESTARTREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NotifyGCSRestartRequest)
  })
_sym_db.RegisterMessage(NotifyGCSRestartRequest)

NotifyGCSRestartReply = _reflection.GeneratedProtocolMessageType('NotifyGCSRestartReply', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFYGCSRESTARTREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NotifyGCSRestartReply)
  })
_sym_db.RegisterMessage(NotifyGCSRestartReply)

GetTaskFailureCauseRequest = _reflection.GeneratedProtocolMessageType('GetTaskFailureCauseRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKFAILURECAUSEREQUEST,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetTaskFailureCauseRequest)
  })
_sym_db.RegisterMessage(GetTaskFailureCauseRequest)

GetTaskFailureCauseReply = _reflection.GeneratedProtocolMessageType('GetTaskFailureCauseReply', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKFAILURECAUSEREPLY,
  '__module__' : 'src.ray.protobuf.node_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetTaskFailureCauseReply)
  })
_sym_db.RegisterMessage(GetTaskFailureCauseReply)

_NODEMANAGERSERVICE = DESCRIPTOR.services_by_name['NodeManagerService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _WORKERBACKLOGREPORT._serialized_start=107
  _WORKERBACKLOGREPORT._serialized_end=219
  _REPORTWORKERBACKLOGREQUEST._serialized_start=222
  _REPORTWORKERBACKLOGREQUEST._serialized_end=350
  _REPORTWORKERBACKLOGREPLY._serialized_start=352
  _REPORTWORKERBACKLOGREPLY._serialized_end=378
  _REQUESTWORKERLEASEREQUEST._serialized_start=381
  _REQUESTWORKERLEASEREQUEST._serialized_end=605
  _REQUESTWORKERLEASEREPLY._serialized_start=608
  _REQUESTWORKERLEASEREPLY._serialized_end=1377
  _REQUESTWORKERLEASEREPLY_SCHEDULINGFAILURETYPE._serialized_start=1139
  _REQUESTWORKERLEASEREPLY_SCHEDULINGFAILURETYPE._serialized_end=1377
  _PREPAREBUNDLERESOURCESREQUEST._serialized_start=1379
  _PREPAREBUNDLERESOURCESREQUEST._serialized_end=1462
  _PREPAREBUNDLERESOURCESREPLY._serialized_start=1464
  _PREPAREBUNDLERESOURCESREPLY._serialized_end=1519
  _COMMITBUNDLERESOURCESREQUEST._serialized_start=1521
  _COMMITBUNDLERESOURCESREQUEST._serialized_end=1603
  _COMMITBUNDLERESOURCESREPLY._serialized_start=1605
  _COMMITBUNDLERESOURCESREPLY._serialized_end=1633
  _CANCELRESOURCERESERVEREQUEST._serialized_start=1635
  _CANCELRESOURCERESERVEREQUEST._serialized_end=1715
  _CANCELRESOURCERESERVEREPLY._serialized_start=1717
  _CANCELRESOURCERESERVEREPLY._serialized_end=1745
  _RETURNWORKERREQUEST._serialized_start=1748
  _RETURNWORKERREQUEST._serialized_end=1915
  _RETURNWORKERREPLY._serialized_start=1917
  _RETURNWORKERREPLY._serialized_end=1936
  _RELEASEUNUSEDWORKERSREQUEST._serialized_start=1938
  _RELEASEUNUSEDWORKERSREQUEST._serialized_end=2010
  _RELEASEUNUSEDWORKERSREPLY._serialized_start=2012
  _RELEASEUNUSEDWORKERSREPLY._serialized_end=2039
  _SHUTDOWNRAYLETREQUEST._serialized_start=2041
  _SHUTDOWNRAYLETREQUEST._serialized_end=2092
  _SHUTDOWNRAYLETREPLY._serialized_start=2094
  _SHUTDOWNRAYLETREPLY._serialized_end=2115
  _CANCELWORKERLEASEREQUEST._serialized_start=2117
  _CANCELWORKERLEASEREQUEST._serialized_end=2168
  _CANCELWORKERLEASEREPLY._serialized_start=2170
  _CANCELWORKERLEASEREPLY._serialized_end=2220
  _PINOBJECTIDSREQUEST._serialized_start=2223
  _PINOBJECTIDSREQUEST._serialized_end=2387
  _PINOBJECTIDSREPLY._serialized_start=2389
  _PINOBJECTIDSREPLY._serialized_end=2438
  _GETNODESTATSREQUEST._serialized_start=2440
  _GETNODESTATSREQUEST._serialized_end=2509
  _OBJECTSTORESTATS._serialized_start=2512
  _OBJECTSTORESTATS._serialized_end=3274
  _GETNODESTATSREPLY._serialized_start=3277
  _GETNODESTATSREPLY._serialized_end=3461
  _GLOBALGCREQUEST._serialized_start=3463
  _GLOBALGCREQUEST._serialized_end=3480
  _GLOBALGCREPLY._serialized_start=3482
  _GLOBALGCREPLY._serialized_end=3497
  _FORMATGLOBALMEMORYINFOREQUEST._serialized_start=3499
  _FORMATGLOBALMEMORYINFOREQUEST._serialized_end=3578
  _FORMATGLOBALMEMORYINFOREPLY._serialized_start=3581
  _FORMATGLOBALMEMORYINFOREPLY._serialized_end=3709
  _REQUESTOBJECTSPILLAGEREQUEST._serialized_start=3711
  _REQUESTOBJECTSPILLAGEREQUEST._serialized_end=3770
  _REQUESTOBJECTSPILLAGEREPLY._serialized_start=3772
  _REQUESTOBJECTSPILLAGEREPLY._serialized_end=3897
  _RELEASEUNUSEDBUNDLESREQUEST._serialized_start=3899
  _RELEASEUNUSEDBUNDLESREQUEST._serialized_end=3983
  _RELEASEUNUSEDBUNDLESREPLY._serialized_start=3985
  _RELEASEUNUSEDBUNDLESREPLY._serialized_end=4012
  _GETSYSTEMCONFIGREQUEST._serialized_start=4014
  _GETSYSTEMCONFIGREQUEST._serialized_end=4038
  _GETSYSTEMCONFIGREPLY._serialized_start=4040
  _GETSYSTEMCONFIGREPLY._serialized_end=4099
  _REQUESTRESOURCEREPORTREQUEST._serialized_start=4101
  _REQUESTRESOURCEREPORTREQUEST._serialized_end=4131
  _REQUESTRESOURCEREPORTREPLY._serialized_start=4133
  _REQUESTRESOURCEREPORTREPLY._serialized_end=4215
  _UPDATERESOURCEUSAGEREQUEST._serialized_start=4217
  _UPDATERESOURCEUSAGEREQUEST._serialized_end=4316
  _UPDATERESOURCEUSAGEREPLY._serialized_start=4318
  _UPDATERESOURCEUSAGEREPLY._serialized_end=4344
  _GETTASKSINFOREQUEST._serialized_start=4346
  _GETTASKSINFOREQUEST._serialized_end=4404
  _GETTASKSINFOREPLY._serialized_start=4407
  _GETTASKSINFOREPLY._serialized_end=4569
  _GETOBJECTSINFOREQUEST._serialized_start=4571
  _GETOBJECTSINFOREQUEST._serialized_end=4631
  _GETOBJECTSINFOREPLY._serialized_start=4633
  _GETOBJECTSINFOREPLY._serialized_end=4748
  _GETRESOURCELOADREQUEST._serialized_start=4750
  _GETRESOURCELOADREQUEST._serialized_end=4774
  _GETRESOURCELOADREPLY._serialized_start=4776
  _GETRESOURCELOADREPLY._serialized_end=4852
  _NOTIFYGCSRESTARTREQUEST._serialized_start=4854
  _NOTIFYGCSRESTARTREQUEST._serialized_end=4879
  _NOTIFYGCSRESTARTREPLY._serialized_start=4881
  _NOTIFYGCSRESTARTREPLY._serialized_end=4904
  _GETTASKFAILURECAUSEREQUEST._serialized_start=4906
  _GETTASKFAILURECAUSEREQUEST._serialized_end=4959
  _GETTASKFAILURECAUSEREPLY._serialized_start=4962
  _GETTASKFAILURECAUSEREPLY._serialized_end=5123
  _NODEMANAGERSERVICE._serialized_start=5126
  _NODEMANAGERSERVICE._serialized_end=7190
# @@protoc_insertion_point(module_scope)



import inspect
import logging
import os
import uuid
from functools import wraps

import ray._private.signature
from ray import Language, cross_language
from ray._private import ray_option_utils
from ray._private.auto_init_hook import auto_init_ray
from ray._private.client_mode_hook import (
    client_mode_convert_function,
    client_mode_should_convert,
)
from ray._private.ray_option_utils import _warn_if_using_deprecated_placement_group
from ray._private.serialization import pickle_dumps
from ray._private.utils import get_runtime_env_info, parse_runtime_env
from ray._raylet import (
    STREAMING_GENERATOR_RETURN,
    PythonFunctionDescriptor,
    StreamingObjectRefGenerator,
)
from ray.util.annotations import DeveloperAPI, PublicAPI
from ray.util.placement_group import _configure_placement_group_based_on_context
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy
from ray.util.tracing.tracing_helper import (
    _inject_tracing_into_function,
    _tracing_task_invocation,
)

logger = logging.getLogger(__name__)


# Hook to call with (fn, resources, strategy) on each local task submission.
_task_launch_hook = None


@PublicAPI
class RemoteFunction:
    """A remote function.

    This is a decorated function. It can be used to spawn tasks.

    Attributes:
        _language: The target language.
        _function: The original function.
        _function_descriptor: The function descriptor. This is not defined
            until the remote function is first invoked because that is when the
            function is pickled, and the pickled function is used to compute
            the function descriptor.
        _function_name: The module and function name.
        _num_cpus: The default number of CPUs to use for invocations of this
            remote function.
        _num_gpus: The default number of GPUs to use for invocations of this
            remote function.
        _memory: The heap memory request in bytes for this task/actor,
            rounded down to the nearest integer.
        _resources: The default custom resource requirements for invocations of
            this remote function.
        _num_returns: The default number of return values for invocations
            of this remote function.
        _max_calls: The number of times a worker can execute this function
            before exiting.
        _max_retries: The number of times this task may be retried
            on worker failure.
        _retry_exceptions: Whether application-level errors should be retried.
            This can be a boolean or a list/tuple of exceptions that should be retried.
        _runtime_env: The runtime environment for this task.
        _decorator: An optional decorator that should be applied to the remote
            function invocation (as opposed to the function execution) before
            invoking the function. The decorator must return a function that
            takes in two arguments ("args" and "kwargs"). In most cases, it
            should call the function that was passed into the decorator and
            return the resulting ObjectRefs. For an example, see
            "test_decorated_function" in "python/ray/tests/test_basic.py".
        _function_signature: The function signature.
        _last_export_session_and_job: A pair of the last exported session
            and job to help us to know whether this function was exported.
            This is an imperfect mechanism used to determine if we need to
            export the remote function again. It is imperfect in the sense that
            the actor class definition could be exported multiple times by
            different workers.
        _scheduling_strategy: Strategy about how to schedule
            this remote function.
    """

    def __init__(
        self,
        language,
        function,
        function_descriptor,
        task_options,
    ):
        if inspect.iscoroutinefunction(function):
            raise ValueError(
                "'async def' should not be used for remote tasks. You can wrap the "
                "async function with `asyncio.run(f())`. See more at:"
                "https://docs.ray.io/en/latest/ray-core/actors/async_api.html "
            )
        self._default_options = task_options

        # When gpu is used, set the task non-recyclable by default.
        # https://github.com/ray-project/ray/issues/29624 for more context.
        num_gpus = self._default_options.get("num_gpus") or 0
        if num_gpus > 0 and self._default_options.get("max_calls", None) is None:
            self._default_options["max_calls"] = 1

        # TODO(suquark): This is a workaround for class attributes of options.
        # They are being used in some other places, mostly tests. Need cleanup later.
        # E.g., actors uses "__ray_metadata__" to collect options, we can so something
        # similar for remote functions.
        for k, v in ray_option_utils.task_options.items():
            setattr(self, "_" + k, task_options.get(k, v.default_value))
        self._runtime_env = parse_runtime_env(self._runtime_env)
        if "runtime_env" in self._default_options:
            self._default_options["runtime_env"] = self._runtime_env

        self._language = language
        self._function = _inject_tracing_into_function(function)
        self._function_name = function.__module__ + "." + function.__name__
        self._function_descriptor = function_descriptor
        self._is_cross_language = language != Language.PYTHON
        self._decorator = getattr(function, "__ray_invocation_decorator__", None)
        self._function_signature = ray._private.signature.extract_signature(
            self._function
        )
        self._last_export_session_and_job = None
        self._uuid = uuid.uuid4()

        # Override task.remote's signature and docstring
        @wraps(function)
        def _remote_proxy(*args, **kwargs):
            return self._remote(args=args, kwargs=kwargs, **self._default_options)

        self.remote = _remote_proxy

    def __call__(self, *args, **kwargs):
        raise TypeError(
            "Remote functions cannot be called directly. Instead "
            f"of running '{self._function_name}()', "
            f"try '{self._function_name}.remote()'."
        )

    def options(self, **task_options):
        """Configures and overrides the task invocation parameters.

        The arguments are the same as those that can be passed to :obj:`ray.remote`.
        Overriding `max_calls` is not supported.

        Args:
            num_returns: It specifies the number of object refs returned by
                the remote function invocation.
            num_cpus: The quantity of CPU cores to reserve
                for this task or for the lifetime of the actor.
            num_gpus: The quantity of GPUs to reserve
                for this task or for the lifetime of the actor.
            resources (Dict[str, float]): The quantity of various custom resources
                to reserve for this task or for the lifetime of the actor.
                This is a dictionary mapping strings (resource names) to floats.
            accelerator_type: If specified, requires that the task or actor run
                on a node with the specified type of accelerator.
                See `ray.util.accelerators` for accelerator types.
            memory: The heap memory request in bytes for this task/actor,
                rounded down to the nearest integer.
            object_store_memory: The object store memory request for actors only.
            max_calls: This specifies the
                maximum number of times that a given worker can execute
                the given remote function before it must exit
                (this can be used to address memory leaks in third-party
                libraries or to reclaim resources that cannot easily be
                released, e.g., GPU memory that was acquired by TensorFlow).
                By default this is infinite for CPU tasks and 1 for GPU tasks
                (to force GPU tasks to release resources after finishing).
            max_retries: This specifies the maximum number of times that the remote
                function should be rerun when the worker process executing it
                crashes unexpectedly. The minimum valid value is 0,
                the default is 3 (default), and a value of -1 indicates
                infinite retries.
            runtime_env (Dict[str, Any]): Specifies the runtime environment for
                this actor or task and its children. See
                :ref:`runtime-environments` for detailed documentation.
            retry_exceptions: This specifies whether application-level errors
                should be retried up to max_retries times.
            scheduling_strategy: Strategy about how to
                schedule a remote function or actor. Possible values are
                None: ray will figure out the scheduling strategy to use, it
                will either be the PlacementGroupSchedulingStrategy using parent's
                placement group if parent has one and has
                placement_group_capture_child_tasks set to true,
                or "DEFAULT";
                "DEFAULT": default hybrid scheduling;
                "SPREAD": best effort spread scheduling;
                `PlacementGroupSchedulingStrategy`:
                placement group based scheduling;
                `NodeAffinitySchedulingStrategy`:
                node id based affinity scheduling.
            _metadata: Extended options for Ray libraries. For example,
                _metadata={"workflows.io/options": <workflow options>} for
                Ray workflows.

        Examples:

        .. code-block:: python

            @ray.remote(num_gpus=1, max_calls=1, num_returns=2)
            def f():
               return 1, 2
            # Task g will require 2 gpus instead of 1.
            g = f.options(num_gpus=2)
        """

        func_cls = self

        # override original options
        default_options = self._default_options.copy()
        # max_calls could not be used in ".options()", we should remove it before
        # merging options from '@ray.remote'.
        default_options.pop("max_calls", None)
        updated_options = ray_option_utils.update_options(default_options, task_options)
        ray_option_utils.validate_task_options(updated_options, in_options=True)

        # only update runtime_env when ".options()" specifies new runtime_env
        if "runtime_env" in task_options:
            updated_options["runtime_env"] = parse_runtime_env(
                updated_options["runtime_env"]
            )

        class FuncWrapper:
            def remote(self, *args, **kwargs):
                return func_cls._remote(args=args, kwargs=kwargs, **updated_options)

            @DeveloperAPI
            def bind(self, *args, **kwargs):
                """
                For Ray DAG building that creates static graph from decorated
                class or functions.
                """
                from ray.dag.function_node import FunctionNode

                return FunctionNode(func_cls._function, args, kwargs, updated_options)

        return FuncWrapper()

    @_tracing_task_invocation
    def _remote(self, args=None, kwargs=None, **task_options):
        """Submit the remote function for execution."""
        # We pop the "max_calls" coming from "@ray.remote" here. We no longer need
        # it in "_remote()".
        task_options.pop("max_calls", None)
        auto_init_ray()
        if client_mode_should_convert():
            return client_mode_convert_function(self, args, kwargs, **task_options)

        worker = ray._private.worker.global_worker
        worker.check_connected()

        # If this function was not exported in this session and job, we need to
        # export this function again, because the current GCS doesn't have it.
        if (
            not self._is_cross_language
            and self._last_export_session_and_job != worker.current_session_and_job
        ):
            self._function_descriptor = PythonFunctionDescriptor.from_function(
                self._function, self._uuid
            )
            # There is an interesting question here. If the remote function is
            # used by a subsequent driver (in the same script), should the
            # second driver pickle the function again? If yes, then the remote
            # function definition can differ in the second driver (e.g., if
            # variables in its closure have changed). We probably want the
            # behavior of the remote function in the second driver to be
            # independent of whether or not the function was invoked by the
            # first driver. This is an argument for repickling the function,
            # which we do here.
            self._pickled_function = pickle_dumps(
                self._function,
                f"Could not serialize the function {self._function_descriptor.repr}",
            )

            self._last_export_session_and_job = worker.current_session_and_job
            worker.function_actor_manager.export(self)

        kwargs = {} if kwargs is None else kwargs
        args = [] if args is None else args

        # fill task required options
        for k, v in ray_option_utils.task_options.items():
            if k == "max_retries":
                # TODO(swang): We need to override max_retries here because the default
                # value gets set at Ray import time. Ideally, we should allow setting
                # default values from env vars for other options too.
                v.default_value = os.environ.get(
                    "RAY_TASK_MAX_RETRIES", v.default_value
                )
                v.default_value = int(v.default_value)
            task_options[k] = task_options.get(k, v.default_value)
        # "max_calls" already takes effects and should not apply again.
        # Remove the default value here.
        task_options.pop("max_calls", None)

        # TODO(suquark): cleanup these fields
        name = task_options["name"]
        runtime_env = parse_runtime_env(task_options["runtime_env"])
        placement_group = task_options["placement_group"]
        placement_group_bundle_index = task_options["placement_group_bundle_index"]
        placement_group_capture_child_tasks = task_options[
            "placement_group_capture_child_tasks"
        ]
        scheduling_strategy = task_options["scheduling_strategy"]
        num_returns = task_options["num_returns"]
        if num_returns == "dynamic":
            num_returns = -1
        elif num_returns == "streaming":
            # TODO(sang): This is a temporary private API.
            # Remove it when we migrate to the streaming generator.
            num_returns = ray._raylet.STREAMING_GENERATOR_RETURN

        max_retries = task_options["max_retries"]
        retry_exceptions = task_options["retry_exceptions"]
        if isinstance(retry_exceptions, (list, tuple)):
            retry_exception_allowlist = tuple(retry_exceptions)
            retry_exceptions = True
        else:
            retry_exception_allowlist = None

        if scheduling_strategy is None or not isinstance(
            scheduling_strategy, PlacementGroupSchedulingStrategy
        ):
            _warn_if_using_deprecated_placement_group(task_options, 4)

        resources = ray._private.utils.resources_from_ray_options(task_options)

        if scheduling_strategy is None or isinstance(
            scheduling_strategy, PlacementGroupSchedulingStrategy
        ):
            if isinstance(scheduling_strategy, PlacementGroupSchedulingStrategy):
                placement_group = scheduling_strategy.placement_group
                placement_group_bundle_index = (
                    scheduling_strategy.placement_group_bundle_index
                )
                placement_group_capture_child_tasks = (
                    scheduling_strategy.placement_group_capture_child_tasks
                )

            if placement_group_capture_child_tasks is None:
                placement_group_capture_child_tasks = (
                    worker.should_capture_child_tasks_in_placement_group
                )
            placement_group = _configure_placement_group_based_on_context(
                placement_group_capture_child_tasks,
                placement_group_bundle_index,
                resources,
                {},  # no placement_resources for tasks
                self._function_descriptor.function_name,
                placement_group=placement_group,
            )
            if not placement_group.is_empty:
                scheduling_strategy = PlacementGroupSchedulingStrategy(
                    placement_group,
                    placement_group_bundle_index,
                    placement_group_capture_child_tasks,
                )
            else:
                scheduling_strategy = "DEFAULT"

        serialized_runtime_env_info = None
        if runtime_env is not None:
            serialized_runtime_env_info = get_runtime_env_info(
                runtime_env,
                is_job_runtime_env=False,
                serialize=True,
            )

        if _task_launch_hook:
            _task_launch_hook(self._function_descriptor, resources, scheduling_strategy)

        def invocation(args, kwargs):
            if self._is_cross_language:
                list_args = cross_language._format_args(worker, args, kwargs)
            elif not args and not kwargs and not self._function_signature:
                list_args = []
            else:
                list_args = ray._private.signature.flatten_args(
                    self._function_signature, args, kwargs
                )

            if worker.mode == ray._private.worker.LOCAL_MODE:
                assert (
                    not self._is_cross_language
                ), "Cross language remote function cannot be executed locally."
            object_refs = worker.core_worker.submit_task(
                self._language,
                self._function_descriptor,
                list_args,
                name if name is not None else "",
                num_returns,
                resources,
                max_retries,
                retry_exceptions,
                retry_exception_allowlist,
                scheduling_strategy,
                worker.debugger_breakpoint,
                serialized_runtime_env_info or "{}",
            )
            # Reset worker's debug context from the last "remote" command
            # (which applies only to this .remote call).
            worker.debugger_breakpoint = b""
            if num_returns == STREAMING_GENERATOR_RETURN:
                # Streaming generator will return a single ref
                # that is for the generator task.
                assert len(object_refs) == 1
                generator_ref = object_refs[0]
                return StreamingObjectRefGenerator(generator_ref, worker)
            if len(object_refs) == 1:
                return object_refs[0]
            elif len(object_refs) > 1:
                return object_refs

        if self._decorator is not None:
            invocation = self._decorator(invocation)

        return invocation(args, kwargs)

    @DeveloperAPI
    def bind(self, *args, **kwargs):
        """
        For Ray DAG building that creates static graph from decorated
        class or functions.
        """

        from ray.dag.function_node import FunctionNode

        return FunctionNode(self._function, args, kwargs, self._default_options)



from enum import Enum
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field
from ray.dashboard.modules.job.common import JobStatus
from ray.util.annotations import PublicAPI


@PublicAPI(stability="beta")
class DriverInfo(BaseModel):
    """A class for recording information about the driver related to the job."""

    id: str = Field(..., description="The id of the driver")
    node_ip_address: str = Field(
        ..., description="The IP address of the node the driver is running on."
    )
    pid: str = Field(
        ..., description="The PID of the worker process the driver is using."
    )
    # TODO(aguo): Add node_id as a field.


@PublicAPI(stability="beta")
class JobType(str, Enum):
    """An enumeration for describing the different job types.

    NOTE:
        This field is still experimental and may change in the future.
    """

    #: A job that was initiated by the Ray Jobs API.
    SUBMISSION = "SUBMISSION"
    #: A job that was initiated by a driver script.
    DRIVER = "DRIVER"


@PublicAPI(stability="beta")
class JobDetails(BaseModel):
    """
    Job data with extra details about its driver and its submission.
    """

    type: JobType = Field(..., description="The type of job.")
    job_id: Optional[str] = Field(
        None,
        description="The job ID. An ID that is created for every job that is "
        "launched in Ray. This can be used to fetch data about jobs using Ray "
        "Core APIs.",
    )
    submission_id: Optional[str] = Field(
        None,
        description="A submission ID is an ID created for every job submitted via"
        "the Ray Jobs API. It can "
        "be used to fetch data about jobs using the Ray Jobs API.",
    )
    driver_info: Optional[DriverInfo] = Field(
        None,
        description="The driver related to this job. For jobs submitted via "
        "the Ray Jobs API, "
        "it is the last driver launched by that job submission, "
        "or None if there is no driver.",
    )

    # The following fields are copied from JobInfo.
    # TODO(aguo): Inherit from JobInfo once it's migrated to pydantic.
    status: JobStatus = Field(..., description="The status of the job.")
    entrypoint: str = Field(..., description="The entrypoint command for this job.")
    message: Optional[str] = Field(
        None, description="A message describing the status in more detail."
    )
    error_type: Optional[str] = Field(
        None, description="Internal error or user script error."
    )
    start_time: Optional[int] = Field(
        None,
        description="The time when the job was started. " "A Unix timestamp in ms.",
    )
    end_time: Optional[int] = Field(
        None,
        description="The time when the job moved into a terminal state. "
        "A Unix timestamp in ms.",
    )
    metadata: Optional[Dict[str, str]] = Field(
        None, description="Arbitrary user-provided metadata for the job."
    )
    runtime_env: Optional[Dict[str, Any]] = Field(
        None, description="The runtime environment for the job."
    )
    # the node info where the driver running on.
    #     - driver_agent_http_address: this node's agent http address
    #     - driver_node_id: this node's id.
    driver_agent_http_address: Optional[str] = Field(
        None,
        description="The HTTP address of the JobAgent on the node the job "
        "entrypoint command is running on.",
    )
    driver_node_id: Optional[str] = Field(
        None,
        description="The node ID of the node the job entrypoint command is running on.",
    )



from collections import defaultdict, deque
from functools import partial
import pathlib
from typing import (
    Any,
    Callable,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    TYPE_CHECKING,
    Union,
)
import uuid

import ray
from ray.rllib.core.learner.reduce_result_dict_fn import _reduce_mean_results
from ray.rllib.core.rl_module.rl_module import (
    ModuleID,
    SingleAgentRLModuleSpec,
    RLMODULE_STATE_DIR_NAME,
)
from ray.rllib.core.learner.learner import LearnerSpec
from ray.rllib.policy.sample_batch import MultiAgentBatch
from ray.rllib.utils.actor_manager import FaultTolerantActorManager
from ray.rllib.utils.minibatch_utils import ShardBatchIterator
from ray.rllib.utils.typing import ResultDict
from ray.rllib.utils.numpy import convert_to_numpy
from ray.train._internal.backend_executor import BackendExecutor
from ray.tune.utils.file_transfer import sync_dir_between_nodes


if TYPE_CHECKING:
    from ray.rllib.core.learner.learner import Learner


def _get_backend_config(learner_class: Type["Learner"]) -> str:
    if learner_class.framework == "torch":
        from ray.train.torch import TorchConfig

        backend_config = TorchConfig()
    elif learner_class.framework == "tf2":
        from ray.train.tensorflow import TensorflowConfig

        backend_config = TensorflowConfig()
    else:
        raise ValueError("framework must be either torch or tf")

    return backend_config


def _is_module_trainable(module_id: ModuleID, batch: MultiAgentBatch) -> bool:
    """Default implemntation for is_module_trainable()

    It assumes that the module is trainable by default.
    """
    return True


class LearnerGroup:
    """Coordinator of Learners.

    Args:
        learner_spec: The specification for constructing Learners.
        max_queue_len: The maximum number of batches to queue up if doing async_update
            If the queue is full itwill evict the oldest batch first.

    """

    def __init__(
        self,
        learner_spec: LearnerSpec,
        max_queue_len: int = 20,
    ):
        scaling_config = learner_spec.learner_group_scaling_config
        learner_class = learner_spec.learner_class

        # TODO (Kourosh): Go with a _remote flag instead of _is_local to be more
        #  explicit.
        self._is_local = scaling_config.num_workers == 0
        self._learner = None
        self._workers = None
        # If a user calls self.shutdown() on their own then this flag is set to true.
        # When del is called the backend executor isn't shutdown twice if this flag is
        # true. the backend executor would otherwise log a warning to the console from
        # ray train.
        self._is_shut_down = False

        self._is_module_trainable = _is_module_trainable

        # How many timesteps had to be dropped due to a full input queue?
        self._in_queue_ts_dropped = 0

        if self._is_local:
            self._learner = learner_class(**learner_spec.get_params_dict())
            self._learner.build()
            self._worker_manager = None
            self._in_queue = []
        else:
            backend_config = _get_backend_config(learner_class)
            backend_executor = BackendExecutor(
                backend_config=backend_config,
                num_workers=scaling_config.num_workers,
                num_cpus_per_worker=scaling_config.num_cpus_per_worker,
                num_gpus_per_worker=scaling_config.num_gpus_per_worker,
                max_retries=0,
            )
            backend_executor.start(
                train_cls=learner_class,
                train_cls_kwargs=learner_spec.get_params_dict(),
            )
            self._backend_executor = backend_executor

            self._workers = [w.actor for w in backend_executor.worker_group.workers]

            # Run the neural network building code on remote workers.
            ray.get([w.build.remote() for w in self._workers])

            self._worker_manager = FaultTolerantActorManager(
                self._workers,
                # TODO (sven): This probably works even without any restriction
                #  (allowing for any arbitrary number of requests in-flight). Test with
                #  3 first, then with unlimited, and if both show the same behavior on
                #  an async algo, remove this restriction entirely.
                max_remote_requests_in_flight_per_actor=3,
            )
            # This is a list of the tags for asynchronous update requests that are
            # inflight, and is used for grouping together the results of requests
            # that were sent to the workers at the same time.
            self._inflight_request_tags: Set[str] = set()
            self._in_queue = deque(maxlen=max_queue_len)

    def get_in_queue_stats(self) -> Mapping[str, Any]:
        """Returns the current stats for the input queue for this learner group."""
        return {
            "learner_group_queue_size": len(self._in_queue),
            "learner_group_queue_ts_dropped": self._in_queue_ts_dropped,
        }

    @property
    def is_local(self) -> bool:
        return self._is_local

    def update(
        self,
        batch: MultiAgentBatch,
        *,
        minibatch_size: Optional[int] = None,
        num_iters: int = 1,
        reduce_fn: Optional[Callable[[List[Mapping[str, Any]]], ResultDict]] = (
            _reduce_mean_results
        ),
    ) -> Union[Mapping[str, Any], List[Mapping[str, Any]]]:
        """Do one or more gradient based updates to the Learner(s) based on given data.

        Args:
            batch: The data batch to use for the update.
            minibatch_size: The minibatch size to use for the update.
            num_iters: The number of complete passes over all the sub-batches in the
                input multi-agent batch.
            reduce_fn: An optional callable to reduce the results from a list of the
                Learner actors into a single result. This can be any arbitrary function
                that takes a list of dictionaries and returns a single dictionary. For
                example you can either take an average (default) or concatenate the
                results (for example for metrics) or be more selective about you want to
                report back to the algorithm's training_step. If None is passed, the
                results will not get reduced.

        Returns:
            A dictionary with the reduced results of the updates from the Learner(s) or
            a list of dictionaries of results from the updates from the Learner(s).
        """

        # Construct a multi-agent batch with only the trainable modules.
        train_batch = {}
        for module_id in batch.policy_batches.keys():
            if self._is_module_trainable(module_id, batch):
                train_batch[module_id] = batch.policy_batches[module_id]
        train_batch = MultiAgentBatch(train_batch, batch.count)

        if self.is_local:
            results = [
                self._learner.update(
                    train_batch,
                    minibatch_size=minibatch_size,
                    num_iters=num_iters,
                    reduce_fn=reduce_fn,
                )
            ]
        else:

            def _learner_update(learner, minibatch):
                return learner.update(
                    minibatch,
                    minibatch_size=minibatch_size,
                    num_iters=num_iters,
                    reduce_fn=reduce_fn,
                )

            results = self._get_results(
                self._worker_manager.foreach_actor(
                    [
                        partial(_learner_update, minibatch=minibatch)
                        for minibatch in ShardBatchIterator(batch, len(self._workers))
                    ]
                )
            )

        # TODO(sven): Move reduce_fn to the training_step
        if reduce_fn is None:
            return results
        else:
            return reduce_fn(results)

    def async_update(
        self,
        batch: MultiAgentBatch,
        *,
        minibatch_size: Optional[int] = None,
        num_iters: int = 1,
        reduce_fn: Optional[Callable[[List[Mapping[str, Any]]], ResultDict]] = (
            _reduce_mean_results
        ),
    ) -> Union[List[Mapping[str, Any]], List[List[Mapping[str, Any]]]]:
        """Asnychronously do gradient based updates to the Learner(s) with `batch`.

        Args:
            batch: The data batch to use for the update.
            minibatch_size: The minibatch size to use for the update.
            num_iters: The number of complete passes over all the sub-batches in the
                input multi-agent batch.
            reduce_fn: An optional callable to reduce the results from a list of the
                Learner actors into a single result. This can be any arbitrary function
                that takes a list of dictionaries and returns a single dictionary. For
                example you can either take an average (default) or concatenate the
                results (for example for metrics) or be more selective about you want to
                report back to the algorithm's training_step. If None is passed, the
                results will not get reduced.

        Returns:
            A list of list of dictionaries of results, where the outer list
            corresponds to separate calls to `async_update`, and the inner
            list corresponds to the results from each Learner(s). Or if the results
            are reduced, a list of dictionaries of the reduced results from each
            call to async_update that is ready.
        """
        if self.is_local:
            raise ValueError(
                "Cannot call `async_update` when running in local mode with "
                "num_workers=0."
            )
        else:
            if minibatch_size is not None:
                minibatch_size //= len(self._workers)

            def _learner_update(learner, minibatch):
                return learner.update(
                    minibatch,
                    minibatch_size=minibatch_size,
                    num_iters=num_iters,
                    reduce_fn=reduce_fn,
                )

            # Queue the new batches.
            # If queue is full, kick out the oldest item (and thus add its
            # length to the "dropped ts" counter).
            if len(self._in_queue) == self._in_queue.maxlen:
                self._in_queue_ts_dropped += len(self._in_queue[0])

            self._in_queue.append(batch)

            # Retrieve all ready results (kicked off by prior calls to this method).
            results = self._worker_manager.fetch_ready_async_reqs(
                tags=list(self._inflight_request_tags)
            )
            # Only if there are no more requests in-flight on any of the learners,
            # we can send in one new batch for sharding and parallel learning.
            if self._worker_manager_ready():
                count = 0
                # TODO (sven): This probably works even without any restriction
                #  (allowing for any arbitrary number of requests in-flight). Test with
                #  3 first, then with unlimited, and if both show the same behavior on
                #  an async algo, remove this restriction entirely.
                while len(self._in_queue) > 0 and count < 3:
                    # Pull a single batch from the queue (from the left side, meaning:
                    # use the oldest one first).
                    update_tag = str(uuid.uuid4())
                    self._inflight_request_tags.add(update_tag)
                    batch = self._in_queue.popleft()
                    self._worker_manager.foreach_actor_async(
                        [
                            partial(_learner_update, minibatch=minibatch)
                            for minibatch in ShardBatchIterator(
                                batch, len(self._workers)
                            )
                        ],
                        tag=update_tag,
                    )
                    count += 1

            # NOTE: There is a strong assumption here that the requests launched to
            # learner workers will return at the same time, since they are have a
            # barrier inside of themselves for gradient aggregation. Therefore results
            # should be a list of lists where each inner list should be the length of
            # the number of learner workers, if results from an  non-blocking update are
            # ready.
            results = self._get_async_results(results)

            # TODO(sven): Move reduce_fn to the training_step
            if reduce_fn is None:
                return results
            else:
                return [reduce_fn(r) for r in results]

    def _worker_manager_ready(self):
        # TODO (sven): This probably works even without any restriction (allowing for
        #  any arbitrary number of requests in-flight). Test with 3 first, then with
        #  unlimited, and if both show the same behavior on an async algo, remove
        #  this method entirely.
        return (
            self._worker_manager.num_outstanding_async_reqs()
            <= self._worker_manager.num_actors() * 2
        )

    def _get_results(self, results):
        processed_results = []
        for result in results:
            result_or_error = result.get()
            if result.ok:
                processed_results.append(result_or_error)
            else:
                raise result_or_error
        return processed_results

    def _get_async_results(self, results):
        """Get results from the worker manager and group them by tag.

        Returns:
            A list of lists of results, where each inner list contains all results
            for same tags.

        """
        unprocessed_results = defaultdict(list)
        for result in results:
            result_or_error = result.get()
            if result.ok:
                assert (
                    result.tag
                ), "Cannot call _get_async_results on untagged async requests."
                unprocessed_results[result.tag].append(result_or_error)
            else:
                raise result_or_error

        for tag in unprocessed_results.keys():
            self._inflight_request_tags.remove(tag)
        return list(unprocessed_results.values())

    def additional_update(
        self,
        *,
        reduce_fn: Callable[[ResultDict], ResultDict] = _reduce_mean_results,
        **kwargs,
    ) -> Union[Mapping[str, Any], List[Mapping[str, Any]]]:
        """Apply additional non-gradient based updates to the Learners.

        For example, this could be used to do a polyak averaging update
        of a target network in off policy algorithms like SAC or DQN.

        By default this is a pass through that calls `Learner.additional_update`

        Args:
            reduce_fn: See `update()` documentation for more details.
            **kwargs: Keyword arguments to pass to each Learner.

        Returns:
            A list of dictionaries of results from the updates from each worker.
        """

        if self.is_local:
            return self._learner.additional_update(**kwargs)
        else:
            results = self._worker_manager.foreach_actor(
                [lambda w: w.additional_update(**kwargs) for _ in self._workers]
            )
            results = self._get_results(results)
            if reduce_fn is None:
                return results
            # TODO(sven): Move reduce_fn to the training_step
            return reduce_fn(results)

    def add_module(
        self,
        *,
        module_id: ModuleID,
        module_spec: SingleAgentRLModuleSpec,
    ) -> None:
        """Add a module to the Learners maintained by this LearnerGroup.

        Args:
            module_id: The id of the module to add.
            module_spec:  #TODO (Kourosh) fill in here.
        """
        if self.is_local:
            self._learner.add_module(
                module_id=module_id,
                module_spec=module_spec,
            )
        else:
            results = self._worker_manager.foreach_actor(
                lambda w: w.add_module(
                    module_id=module_id,
                    module_spec=module_spec,
                )
            )
            return self._get_results(results)

    def remove_module(self, module_id: ModuleID) -> None:
        """Remove a module from the Learners maintained by this LearnerGroup.

        Args:
            module_id: The id of the module to remove.

        """
        if self.is_local:
            self._learner.remove_module(module_id)
        else:
            refs = []
            for worker in self._workers:
                ref = worker.remove_module.remote(module_id)
                refs.append(ref)
            ray.get(refs)

    def set_weights(self, weights: Mapping[str, Any]) -> None:
        """Set the weights of the MultiAgentRLModule maintained by each Learner.

        The weights don't have to include all the modules in the MARLModule.
            This way the weights of only some of the Agents can be set.

        Args:
            weights: The weights to set each RLModule in the MARLModule to.

        """
        if self.is_local:
            self._learner.set_module_state(weights)
        else:
            results_or_errors = self._worker_manager.foreach_actor(
                lambda w: w.set_module_state(weights)
            )
            # raise errors if any
            self._get_results(results_or_errors)

    def get_weights(self, module_ids: Optional[Set[str]] = None) -> Mapping[str, Any]:
        """Get the weights of the MultiAgentRLModule maintained by each Learner.

        Args:
            module_ids: The ids of the modules to get the weights of.

        Returns:
            A mapping of module ids to their weights.

        """
        if self.is_local:
            state = self._learner.get_module_state(module_ids)
        else:
            worker = self._worker_manager.healthy_actor_ids()[0]
            assert len(self._workers) == self._worker_manager.num_healthy_actors()
            state = self._worker_manager.foreach_actor(
                lambda w: w.get_module_state(module_ids), remote_actor_ids=[worker]
            )
            state = self._get_results(state)[0]

        return convert_to_numpy(state)

    def get_state(self) -> Mapping[ModuleID, Mapping[str, Any]]:
        """Get the states of the first Learners.

        This should be the same across Learners
        """
        if self.is_local:
            return self._learner.get_state()
        else:
            worker = self._worker_manager.healthy_actor_ids()[0]
            assert len(self._workers) == self._worker_manager.num_healthy_actors()
            results = self._worker_manager.foreach_actor(
                lambda w: w.get_state(), remote_actor_ids=[worker]
            )
            return self._get_results(results)[0]

    def set_state(self, state: List[Mapping[ModuleID, Mapping[str, Any]]]) -> None:
        """Sets the states of the Learners.

        Args:
            state: The state of the Learners

        """
        if self.is_local:
            self._learner.set_state(state)
        else:
            self._worker_manager.foreach_actor(lambda w: w.set_state(state))

    def set_is_module_trainable(
        self, is_module_trainable: Callable[[ModuleID, MultiAgentBatch], bool] = None
    ) -> None:
        """Sets the function that determines whether a module is trainable.

        Args:
            is_module_trainable: A function that takes in a module id and a batch
                and returns a boolean indicating whether the module should be trained
                on the batch.
        """
        if is_module_trainable is not None:
            self._is_module_trainable = is_module_trainable

    def save_state(self, path: str) -> None:
        """Saves the state of the LearnerGroup.

        Args:
            path: The path to save the state to.
        """
        if self.is_local:
            self._learner.save_state(path)
        else:
            worker = self._worker_manager.healthy_actor_ids()[0]
            worker_ip_addr = self._worker_manager.foreach_actor(
                self._get_ip_address, remote_actor_ids=[worker]
            )
            worker_ip_addr = self._get_results(worker_ip_addr)[0]
            self_ip_addr = self._get_ip_address()

            if worker_ip_addr == self_ip_addr:
                self._worker_manager.foreach_actor(
                    lambda w: w.save_state(path), remote_actor_ids=[worker]
                )
            else:
                # save the checkpoint to a temporary location on the worker

                # create a temporary directory on the worker
                worker_temp_dir = self._worker_manager.foreach_actor(
                    self._create_temporary_dir, remote_actor_ids=[worker]
                )
                worker_temp_dir = self._get_results(worker_temp_dir)[0]

                # save the checkpoint to the temporary directory on the worker
                self._worker_manager.foreach_actor(
                    lambda w: w.save_state(worker_temp_dir), remote_actor_ids=[worker]
                )

                # sync the temporary directory on the worker to the local directory
                sync_dir_between_nodes(
                    worker_ip_addr, worker_temp_dir, self_ip_addr, path
                )

                # creating this function here instead of making it a member funciton
                # becasue it uses the worker_temp_dir variable, and this can't
                # be passed in as an argument to foreach_actor
                def remove_dir(w):
                    import shutil

                    shutil.rmtree(worker_temp_dir)

                # remove the temporary directory on the worker
                self._worker_manager.foreach_actor(
                    remove_dir, remote_actor_ids=[worker]
                )

    def load_state(self, path: str) -> None:
        """Loads the state of the LearnerGroup.

        Args:
            path: The path to load the state from.
        """
        path = str(self._resolve_checkpoint_path(path))

        if self.is_local:
            self._learner.load_state(path)
        else:
            assert len(self._workers) == self._worker_manager.num_healthy_actors()
            head_node_ip = ray.util.get_node_ip_address()
            workers = self._worker_manager.healthy_actor_ids()

            def _load_state(w):
                # doing imports here since they might not be imported on the worker
                import ray
                import tempfile

                worker_node_ip = ray.util.get_node_ip_address()
                # if the worker is on the same node as the head, load the checkpoint
                # directly from the path otherwise sync the checkpoint from the head
                # to the worker and load it from there
                if worker_node_ip == head_node_ip:
                    w.load_state(path)
                else:
                    with tempfile.TemporaryDirectory() as temp_dir:
                        sync_dir_between_nodes(
                            head_node_ip, path, worker_node_ip, temp_dir
                        )
                        w.load_state(temp_dir)

            self._worker_manager.foreach_actor(_load_state, remote_actor_ids=workers)

    def load_module_state(
        self,
        *,
        marl_module_ckpt_dir: Optional[str] = None,
        modules_to_load: Optional[Set[str]] = None,
        rl_module_ckpt_dirs: Optional[Mapping[ModuleID, str]] = None,
    ) -> None:

        """Load the checkpoints of the modules being trained by this LearnerGroup.

        `load_module_state` can be used 3 ways:
            1. Load a checkpoint for the MultiAgentRLModule being trained by this
                LearnerGroup. Limit the modules that are loaded from the checkpoint
                by specifying the `modules_to_load` argument.
            2. Load the checkpoint(s) for single agent RLModules that
                are in the MultiAgentRLModule being trained by this LearnerGroup.
            3. Load a checkpoint for the MultiAgentRLModule being trained by this
                LearnerGroup and load the checkpoint(s) for single agent RLModules
                that are in the MultiAgentRLModule. The checkpoints for the single
                agent RLModules take precedence over the module states in the
                MultiAgentRLModule checkpoint.

        NOTE: At lease one of marl_module_ckpt_dir or rl_module_ckpt_dirs is
            must be specified. modules_to_load can only be specified if
            marl_module_ckpt_dir is specified.

        Args:
            marl_module_ckpt_dir: The path to the checkpoint for the
                MultiAgentRLModule.
            modules_to_load: A set of module ids to load from the checkpoint.
            rl_module_ckpt_dirs: A mapping from module ids to the path to a
                checkpoint for a single agent RLModule.
        """
        if not (marl_module_ckpt_dir or rl_module_ckpt_dirs):
            raise ValueError(
                "At least one of multi_agent_module_state or "
                "single_agent_module_states must be specified."
            )
        if marl_module_ckpt_dir:
            if not isinstance(marl_module_ckpt_dir, str):
                raise ValueError("multi_agent_module_state must be a string path.")
            marl_module_ckpt_dir = self._resolve_checkpoint_path(marl_module_ckpt_dir)
        if rl_module_ckpt_dirs:
            if not isinstance(rl_module_ckpt_dirs, dict):
                raise ValueError("single_agent_module_states must be a dictionary.")
            for module_id, path in rl_module_ckpt_dirs.items():
                if not isinstance(path, str):
                    raise ValueError(
                        "rl_module_ckpt_dirs must be a dictionary "
                        "mapping module ids to string paths."
                    )
                rl_module_ckpt_dirs[module_id] = self._resolve_checkpoint_path(path)
        if modules_to_load:
            if not isinstance(modules_to_load, set):
                raise ValueError("modules_to_load must be a set.")
            for module_id in modules_to_load:
                if not isinstance(module_id, str):
                    raise ValueError("modules_to_load must be a list of strings.")

        if self.is_local:
            module_keys = set(self._learner.module.keys())
        else:
            workers = self._worker_manager.healthy_actor_ids()
            module_keys = set(
                self._get_results(
                    self._worker_manager.foreach_actor(
                        lambda w: w.module.keys(), remote_actor_ids=[workers[0]]
                    )
                )[0]
            )

        if marl_module_ckpt_dir and rl_module_ckpt_dirs:
            # If both a MARLModule checkpoint and RLModule checkpoints are specified,
            # the RLModule checkpoints take precedence over the MARLModule checkpoint,
            # so we should not load any modules in the MARLModule checkpoint that are
            # also in the RLModule checkpoints.
            if modules_to_load:
                if any(
                    module_id in modules_to_load
                    for module_id in rl_module_ckpt_dirs.keys()
                ):
                    raise ValueError(
                        f"module_id {module_id} was specified in both "
                        "modules_to_load and rl_module_ckpt_dirs. Please only "
                        "specify a module to be loaded only once, either in "
                        "modules_to_load or rl_module_ckpt_dirs, but not both."
                    )
            else:
                modules_to_load = module_keys - set(rl_module_ckpt_dirs.keys())

        # No need to do any file transfer operations if we are running training
        # on the experiment head node.
        if self._is_local:
            if marl_module_ckpt_dir:
                # load the MARLModule checkpoint if they were specified
                self._learner.module.load_state(
                    marl_module_ckpt_dir, modules_to_load=modules_to_load
                )
            if rl_module_ckpt_dirs:
                # load the RLModule if they were specified
                for module_id, path in rl_module_ckpt_dirs.items():
                    self._learner.module[module_id].load_state(
                        path / RLMODULE_STATE_DIR_NAME
                    )
        else:
            self._distributed_load_module_state(
                marl_module_ckpt_dir=marl_module_ckpt_dir,
                modules_to_load=modules_to_load,
                rl_module_ckpt_dirs=rl_module_ckpt_dirs,
            )

    def _distributed_load_module_state(
        self,
        *,
        marl_module_ckpt_dir: Optional[str] = None,
        modules_to_load: Optional[Set[str]] = None,
        rl_module_ckpt_dirs: Optional[Mapping[ModuleID, str]] = None,
    ):
        """Load the checkpoints of the modules being trained by this LearnerGroup.

           This method only needs to be called if the LearnerGroup is training
           distributed learners (e.g num_learner_workers > 0).

        Args:
            marl_module_ckpt_dir: The path to the checkpoint for the
                MultiAgentRLModule.
            modules_to_load: A set of module ids to load from the checkpoint.
            rl_module_ckpt_dirs: A mapping from module ids to the path to a
                checkpoint for a single agent RLModule.

        """

        assert len(self._workers) == self._worker_manager.num_healthy_actors()
        workers = self._worker_manager.healthy_actor_ids()
        head_node_ip = ray.util.get_node_ip_address()

        def _load_module_state(w):
            # doing imports here since they might not be imported on the worker
            import ray
            import tempfile
            import shutil

            worker_node_ip = ray.util.get_node_ip_address()
            # sync the checkpoints from the head to the worker if the worker is not
            # on the same node as the head
            tmp_marl_module_ckpt_dir = marl_module_ckpt_dir
            tmp_rl_module_ckpt_dirs = rl_module_ckpt_dirs
            if worker_node_ip != head_node_ip:
                if marl_module_ckpt_dir:
                    tmp_marl_module_ckpt_dir = tempfile.mkdtemp()
                    sync_dir_between_nodes(
                        source_ip=head_node_ip,
                        source_path=marl_module_ckpt_dir,
                        target_ip=worker_node_ip,
                        target_path=tmp_marl_module_ckpt_dir,
                    )
                if rl_module_ckpt_dirs:
                    tmp_rl_module_ckpt_dirs = {}
                    for module_id, path in rl_module_ckpt_dirs.items():
                        tmp_rl_module_ckpt_dirs[module_id] = tempfile.mkdtemp()
                        sync_dir_between_nodes(
                            source_ip=head_node_ip,
                            source_path=path,
                            target_ip=worker_node_ip,
                            target_path=tmp_rl_module_ckpt_dirs[module_id],
                        )
                        tmp_rl_module_ckpt_dirs[module_id] = pathlib.Path(
                            tmp_rl_module_ckpt_dirs[module_id]
                        )
            if marl_module_ckpt_dir:
                # load the MARLModule checkpoint if they were specified
                w.module.load_state(
                    tmp_marl_module_ckpt_dir, modules_to_load=modules_to_load
                )
            if rl_module_ckpt_dirs:
                # load the RLModule if they were specified
                for module_id, path in tmp_rl_module_ckpt_dirs.items():
                    w.module[module_id].load_state(path / RLMODULE_STATE_DIR_NAME)

            # remove the temporary directories on the worker if any were created
            if worker_node_ip != head_node_ip:
                if marl_module_ckpt_dir:
                    shutil.rmtree(tmp_marl_module_ckpt_dir)
                if rl_module_ckpt_dirs:
                    for module_id, path in tmp_rl_module_ckpt_dirs.items():
                        shutil.rmtree(path)

        self._worker_manager.foreach_actor(_load_module_state, remote_actor_ids=workers)

    @staticmethod
    def _resolve_checkpoint_path(path: str) -> pathlib.Path:
        """Checks that the provided checkpoint path is a dir and makes it absolute."""
        path = pathlib.Path(path)
        if not path.is_dir():
            raise ValueError(
                f"Path {path} is not a directory. "
                "Please specify a directory containing the checkpoint files."
            )
        if not path.exists():
            raise ValueError(f"Path {path} does not exist.")
        path = path.absolute()
        return path

    @staticmethod
    def _create_temporary_dir(_=None) -> str:
        """Creates a temporary directory.

        Args:
            _: Unused arg. Exists to make this function compatible with foreach_actor
            calls.

        Returns:
            The path to the temporary directory.
        """
        import tempfile

        return tempfile.mkdtemp()

    @staticmethod
    def _get_ip_address(_=None) -> str:
        """Returns this process's address.

        Args:
            _: Unused arg. Exists to make this function compatible with foreach_actor
            calls.

        Returns:
            The address of this process.

        """
        import ray

        return ray.util.get_node_ip_address()

    def shutdown(self):
        """Shuts down the LearnerGroup."""
        if not self._is_local:
            self._backend_executor.shutdown()
            self._is_shut_down = True

    def __del__(self):
        if not self._is_shut_down:
            self.shutdown()



"""
[1] Mastering Diverse Domains through World Models - 2023
D. Hafner, J. Pasukonis, J. Ba, T. Lillicrap
https://arxiv.org/pdf/2301.04104v1.pdf
"""
from typing import Optional

from ray.rllib.algorithms.dreamerv3.tf.models.components.mlp import MLP
from ray.rllib.algorithms.dreamerv3.tf.models.components.reward_predictor_layer import (
    RewardPredictorLayer,
)
from ray.rllib.utils.framework import try_import_tf

_, tf, _ = try_import_tf()


class CriticNetwork(tf.keras.Model):
    """The critic network described in [1], predicting values for policy learning.

    Contains a copy of itself (EMA net) for weight regularization.
    The EMA net is updated after each train step via EMA (using the `ema_decay`
    parameter and the actual critic's weights). The EMA net is NOT used for target
    computations (we use the actual critic for that), its only purpose is to compute a
    weights regularizer term for the critic's loss such that the actual critic does not
    move too quickly.
    """

    def __init__(
        self,
        *,
        model_size: Optional[str] = "XS",
        num_buckets: int = 255,
        lower_bound: float = -20.0,
        upper_bound: float = 20.0,
        ema_decay: float = 0.98,
    ):
        """Initializes a CriticNetwork instance.

        Args:
            model_size: The "Model Size" used according to [1] Appendinx B.
               Use None for manually setting the different network sizes.
            num_buckets: The number of buckets to create. Note that the number of
                possible symlog'd outcomes from the used distribution is
                `num_buckets` + 1:
                lower_bound --bucket-- o[1] --bucket-- o[2] ... --bucket-- upper_bound
                o=outcomes
                lower_bound=o[0]
                upper_bound=o[num_buckets]
            lower_bound: The symlog'd lower bound for a possible reward value.
                Note that a value of -20.0 here already allows individual (actual env)
                rewards to be as low as -400M. Buckets will be created between
                `lower_bound` and `upper_bound`.
            upper_bound: The symlog'd upper bound for a possible reward value.
                Note that a value of +20.0 here already allows individual (actual env)
                rewards to be as high as 400M. Buckets will be created between
                `lower_bound` and `upper_bound`.
            ema_decay: The weight to use for updating the weights of the critic's copy
                vs the actual critic. After each training update, the EMA copy of the
                critic gets updated according to:
                ema_net=(`ema_decay`*ema_net) + (1.0-`ema_decay`)*critic_net
                The EMA copy of the critic is used inside the critic loss function only
                to produce a regularizer term against the current critic's weights, NOT
                to compute any target values.
        """
        super().__init__(name="critic")

        self.model_size = model_size
        self.ema_decay = ema_decay

        # "Fast" critic network(s) (mlp + reward-pred-layer). This is the network
        # we actually train with our critic loss.
        # IMPORTANT: We also use this to compute the return-targets, BUT we regularize
        # the critic loss term such that the weights of this fast critic stay close
        # to the EMA weights (see below).
        self.mlp = MLP(
            model_size=self.model_size,
            output_layer_size=None,
        )
        self.return_layer = RewardPredictorLayer(
            num_buckets=num_buckets,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
        )

        # Weights-EMA (EWMA) containing networks for critic loss (similar to a
        # target net, BUT not used to compute anything, just for the
        # weights regularizer term inside the critic loss).
        self.mlp_ema = MLP(
            model_size=self.model_size,
            output_layer_size=None,
            trainable=False,
        )
        self.return_layer_ema = RewardPredictorLayer(
            num_buckets=num_buckets,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            trainable=False,
        )

    def call(self, h, z, return_logits=False, use_ema=False):
        """Performs a forward pass through the critic network.

        Args:
            h: The deterministic hidden state of the sequence model. [B, dim(h)].
            z: The stochastic discrete representations of the original
                observation input. [B, num_categoricals, num_classes].
            return_logits: Whether also return (as a second tuple item) the logits
                computed by the binned return layer (instead of only the value itself).
            use_ema: Whether to use the EMA-copy of the critic instead of the actual
                critic to perform this computation.
        """
        # Flatten last two dims of z.
        assert len(z.shape) == 3
        z_shape = tf.shape(z)
        z = tf.reshape(tf.cast(z, tf.float32), shape=(z_shape[0], -1))
        assert len(z.shape) == 2
        out = tf.concat([h, z], axis=-1)

        if not use_ema:
            # Send h-cat-z through MLP.
            out = self.mlp(out)
            # Return expected return OR (expected return, probs of bucket values).
            return self.return_layer(out, return_logits=return_logits)
        else:
            out = self.mlp_ema(out)
            return self.return_layer_ema(out, return_logits=return_logits)

    def init_ema(self) -> None:
        """Initializes the EMA-copy of the critic from the critic's weights.

        After calling this method, the two networks have identical weights.
        """
        vars = self.mlp.trainable_variables + self.return_layer.trainable_variables
        vars_ema = self.mlp_ema.variables + self.return_layer_ema.variables
        assert len(vars) == len(vars_ema)
        for var, var_ema in zip(vars, vars_ema):
            var_ema.assign(var)

    def update_ema(self) -> None:
        """Updates the EMA-copy of the critic according to the update formula:

        ema_net=(`ema_decay`*ema_net) + (1.0-`ema_decay`)*critic_net
        """
        vars = self.mlp.trainable_variables + self.return_layer.trainable_variables
        vars_ema = self.mlp_ema.variables + self.return_layer_ema.variables
        assert len(vars) == len(vars_ema)
        for var, var_ema in zip(vars, vars_ema):
            var_ema.assign(self.ema_decay * var_ema + (1.0 - self.ema_decay) * var)



"""An example of implementing a centralized critic with ObservationFunction.

The advantage of this approach is that it's very simple and you don't have to
change the algorithm at all -- just use callbacks and a custom model.
However, it is a bit less principled in that you have to change the agent
observation spaces to include data that is only used at train time.

See also: centralized_critic.py for an alternative approach that instead
modifies the policy to add a centralized value function.
"""

import numpy as np
from gymnasium.spaces import Dict, Discrete
import argparse
import os

from ray import air, tune
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.algorithms.ppo import PPOConfig
from ray.rllib.examples.models.centralized_critic_models import (
    YetAnotherCentralizedCriticModel,
    YetAnotherTorchCentralizedCriticModel,
)
from ray.rllib.examples.env.two_step_game import TwoStepGame
from ray.rllib.models import ModelCatalog
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.test_utils import check_learning_achieved

parser = argparse.ArgumentParser()
parser.add_argument(
    "--framework",
    choices=["tf", "tf2", "torch"],
    default="torch",
    help="The DL framework specifier.",
)
parser.add_argument(
    "--as-test",
    action="store_true",
    help="Whether this script should be run as a test: --stop-reward must "
    "be achieved within --stop-timesteps AND --stop-iters.",
)
parser.add_argument(
    "--stop-iters", type=int, default=100, help="Number of iterations to train."
)
parser.add_argument(
    "--stop-timesteps", type=int, default=100000, help="Number of timesteps to train."
)
parser.add_argument(
    "--stop-reward", type=float, default=7.99, help="Reward at which we stop training."
)


class FillInActions(DefaultCallbacks):
    """Fills in the opponent actions info in the training batches."""

    def on_postprocess_trajectory(
        self,
        worker,
        episode,
        agent_id,
        policy_id,
        policies,
        postprocessed_batch,
        original_batches,
        **kwargs
    ):
        to_update = postprocessed_batch[SampleBatch.CUR_OBS]
        other_id = 1 if agent_id == 0 else 0
        action_encoder = ModelCatalog.get_preprocessor_for_space(Discrete(2))

        # set the opponent actions into the observation
        _, opponent_batch = original_batches[other_id]
        opponent_actions = np.array(
            [action_encoder.transform(a) for a in opponent_batch[SampleBatch.ACTIONS]]
        )
        to_update[:, -2:] = opponent_actions


def central_critic_observer(agent_obs, **kw):
    """Rewrites the agent obs to include opponent data for training."""

    new_obs = {
        0: {
            "own_obs": agent_obs[0],
            "opponent_obs": agent_obs[1],
            "opponent_action": 0,  # filled in by FillInActions
        },
        1: {
            "own_obs": agent_obs[1],
            "opponent_obs": agent_obs[0],
            "opponent_action": 0,  # filled in by FillInActions
        },
    }
    return new_obs


if __name__ == "__main__":
    args = parser.parse_args()

    ModelCatalog.register_custom_model(
        "cc_model",
        YetAnotherTorchCentralizedCriticModel
        if args.framework == "torch"
        else YetAnotherCentralizedCriticModel,
    )

    action_space = Discrete(2)
    observer_space = Dict(
        {
            "own_obs": Discrete(6),
            # These two fields are filled in by the CentralCriticObserver, and are
            # not used for inference, only for training.
            "opponent_obs": Discrete(6),
            "opponent_action": Discrete(2),
        }
    )

    config = (
        PPOConfig()
        .environment(TwoStepGame)
        .framework(args.framework)
        .rollouts(
            batch_mode="complete_episodes",
            num_rollout_workers=0,
            # TODO(avnishn) make a new example compatible w connectors.
            enable_connectors=False,
        )
        .callbacks(FillInActions)
        # TODO (Kourosh): Lift this example to the new RLModule stack, and enable it.
        .training(model={"custom_model": "cc_model"}, _enable_learner_api=False)
        .multi_agent(
            policies={
                "pol1": (None, observer_space, action_space, {}),
                "pol2": (None, observer_space, action_space, {}),
            },
            policy_mapping_fn=lambda agent_id, episode, worker, **kwargs: "pol1"
            if agent_id == 0
            else "pol2",
            observation_fn=central_critic_observer,
        )
        # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
        .resources(num_gpus=int(os.environ.get("RLLIB_NUM_GPUS", "0")))
        .rl_module(_enable_rl_module_api=False)
    )

    stop = {
        "training_iteration": args.stop_iters,
        "timesteps_total": args.stop_timesteps,
        "episode_reward_mean": args.stop_reward,
    }

    tuner = tune.Tuner(
        "PPO",
        param_space=config.to_dict(),
        run_config=air.RunConfig(stop=stop, verbose=1),
    )
    results = tuner.fit()

    if args.as_test:
        check_learning_achieved(results, args.stop_reward)
